"""
converter.py
====================================
Module for the converter functionality.
"""

from numbers import Number
from pathlib import Path
from sys import stdout
from typing import Union, Optional

import numpy as np
import pandas as pd
from tqdm import tqdm

from gaiaxpy.core.generic_functions import cast_output, validate_pwl_sampling, format_sampled_output
from gaiaxpy.core.generic_variables import pbar_colour, pbar_units, pbar_message
from gaiaxpy.core.satellite import BANDS
from gaiaxpy.input_reader.input_reader import InputReader
from gaiaxpy.output.sampled_spectra_data import SampledSpectraData
from gaiaxpy.spectrum.sampled_basis_functions import SampledBasisFunctions
from gaiaxpy.spectrum.xp_continuous_spectrum import XpContinuousSpectrum
from gaiaxpy.spectrum.xp_sampled_spectrum import XpSampledSpectrum
from .config import parse_config, get_bands_config
from ..config.paths import hermite_bases_file
from ..core.input_validator import validate_save_arguments

__FUNCTION_KEY = 'converter'


def convert(input_object: Union[list, Path, pd.DataFrame, str],
            sampling: Optional[np.ndarray] = np.linspace(0, 60, 600),
            truncation: bool = False, with_correlation: bool = False, output_path: Union[Path, str] = '.',
            output_file: str = 'output_spectra', output_format: str = None, save_file: bool = True,
            username: str = None, password: str = None) -> (pd.DataFrame, np.ndarray):
    """
    Conversion utility: converts the input internally calibrated mean spectra from the continuous representation to a
        sampled form. The sampling grid can be defined by the user, alternatively a default will be adopted. Optionally,
        the continuous representation can be truncated dropping the bases functions (and corresponding coefficients)
        that were considered not to be significant considering the errors on the reconstructed mean spectra.

    Args:
        input_object (list/Path/pd.DataFrame/str): Path to the file containing the mean spectra as downloaded from the
            Archive in their continuous representation, a list of sources ids (string or long), or a pandas DataFrame.
        sampling (ndarray): 1D array containing the desired sampling in pseudo-wavelengths.
        truncation (bool): Toggle truncation of the set of bases. The level of truncation to be applied is defined by
            the recommended value in the input files.
        with_correlation (bool): Whether correlation information should be generated.
        output_path (Path/str): Path where to save the output data.
        output_file (str): Name of the output file without extension (e.g. 'my_file').
        output_format (str): Desired output format. If no format is given, the output file format will be the same as
            the input file (e.g. 'csv').
        save_file (bool): Whether to save the output in a file. If false, output_format and output_file will be ignored.
        username (str): Cosmos username, only suggested when input_object is a list or ADQL query.
        password (str): Cosmos password, only suggested when input_object is a list or ADQL query.

    Returns:
        (tuple): tuple containing:
            DataFrame: The values for all sampled spectra.
            ndarray: The sampling used to convert the input spectra (user-provided or default).

    Raises:
        ValueError: If the sampling is out of the expected boundaries.
    """
    return _convert(input_object=input_object, sampling=sampling, truncation=truncation,
                    with_correlation=with_correlation, output_path=output_path, output_file=output_file,
                    output_format=output_format, save_file=save_file, username=username, password=password)


def _convert(input_object: Union[list, Path, str], sampling: np.ndarray = np.linspace(0, 60, 600),
             truncation: bool = False, with_correlation: bool = False, output_path: Union[Path, str] = '.',
             output_file: str = 'output_spectra', output_format: str = None, save_file: bool = True,
             username: str = None, password: str = None, disable_info: bool = False, config_file=hermite_bases_file) \
        -> (pd.DataFrame, np.ndarray):
    """
    Internal method of the calibration utility. Refer to "convert".

    Args:
        disable_info (bool): Whether to disable the progress tracker.

    Returns:
        DataFrame: A list of all sampled absolute spectra.
        ndarray: The sampling used to calibrate the spectra.

    Raises:
        ValueError: If the sampling is out of the expected boundaries.
    """
    function = convert
    validate_pwl_sampling(sampling)
    validate_save_arguments(function.__defaults__[4], output_file, function.__defaults__[5], output_format, save_file)
    parsed_input_data, extension = InputReader(input_object, convert, truncation=truncation, disable_info=disable_info,
                                               user=username, password=password).read()
    bases_config = parse_config(config_file)
    design_matrices = get_design_matrices(sampling, bases_config)
    spectra_df, positions = _create_spectra(parsed_input_data, truncation, design_matrices,
                                            with_correlation=with_correlation, disable_info=disable_info)
    # Save output section
    output_data = SampledSpectraData(spectra_df, positions)
    output_data.data = cast_output(output_data)
    output_data.save(save_file, output_path, output_file, output_format, extension)
    return output_data.data, positions


def _create_spectrum(row: pd.Series, truncation: bool, design_matrices: dict, band: str,
                     with_correlation: bool = False) -> XpSampledSpectrum:
    """
    Create a single sampled spectrum from the input continuously-represented mean spectrum and design matrix.

    Args:
        row (pd.Series): Single row in a DataFrame containing the entry for one source in the mean spectra file.
            This will include columns for both bands (although one could be missing).
        truncation (bool): Toggle truncation of the set of bases. The level of truncation to be applied is defined by
            the recommended value in the input files.
        design_matrices (dict): 2D array containing the basis functions sampled on the pseudo-wavelength grid (either
            user-defined or default).
        band (str): bp/rp band.
        with_correlation (bool): Whether correlation information should be generated.

    Returns:
        XpSampledSpectrum: The sampled spectrum.
    """
    recommended_truncation = row[f'{band}_n_relevant_bases'] if truncation else -1
    continuous_spectrum = XpContinuousSpectrum(row['source_id'], band, row[f'{band}_coefficients'],
                                               row[f'{band}_covariance_matrix'], row[f'{band}_standard_deviation'])
    return XpSampledSpectrum.from_continuous(continuous_spectrum, design_matrices.get(band),
                                             truncation=recommended_truncation, with_correlation=with_correlation)


def _create_spectra(parsed_input_data: pd.DataFrame, truncation: bool, design_matrices: dict,
                    with_correlation: bool = False, disable_info=False) -> tuple:
    """
    Creates a spectra dataframe from parsed input data and given parameters.

    Args:
        parsed_input_data (pd.DataFrame): The parsed input data to create the spectra from.
        truncation (bool): Toggle truncation of the set of bases. The level of truncation to be applied is defined by
            the recommended value in the input files.
        design_matrices (dict): The design matrices for the input list of bases.
        with_correlation (bool): Whether to include the covariance matrix in the spectra. Default is False.

    Returns:
        (tuple): tuple containing:
            DataFrame: The output spectra.
            ndarray: The sampling used to convert the input spectra (user-provided or default).
    """

    def create_xp_spectra(row, _truncation, _design_matrices, _with_correlation=False):
        """
        Creates bp and rp spectra for a single row of parsed input data.

        Args:
            row (pandas Series): A single row of parsed input data.
            _truncation (bool): Toggle truncation of the set of bases. The level of truncation to be applied is defined
                by the recommended value in the input files.
            _design_matrices (dict): The design matrices for the input list of bases.
            _with_correlation (bool): Whether to include the covariance matrix in the spectra. Default is False.

        Returns:
            list: A list of spectra for the given row of parsed input data containing one element per band available.
        """
        return [_create_spectrum(row, _truncation, _design_matrices, band, with_correlation=_with_correlation)
                for band in BANDS]

    parsed_input_data_dict = parsed_input_data.to_dict('records')
    spectra_series = pd.Series([create_xp_spectra(row, truncation, design_matrices, with_correlation)
                                for row in tqdm(parsed_input_data_dict, desc=pbar_message[__FUNCTION_KEY],
                                                unit=pbar_units[__FUNCTION_KEY], leave=False, colour=pbar_colour,
                                                disable=disable_info, file=stdout)])
    spectra_series = spectra_series.explode()
    return format_sampled_output(spectra_series, with_correlation=with_correlation)


def get_unique_basis_ids(parsed_input_data: pd.DataFrame) -> set:
    """
    Get the IDs of the unique basis required to sample all spectra in the input files.

    Args:
        parsed_input_data (DataFrame): Pandas DataFrame populated with the content of the file containing the mean
            spectra in continuous representation.

    Returns:
        set: A set containing all the required unique basis function IDs.
    """

    # Keep only non-NaN values (in Python, nan != nan)
    def remove_nans(_set):
        return {int(element) for element in _set if element == element}

    set_bp = set([basis for basis in parsed_input_data[f'{BANDS.bp}_basis_function_id'] if isinstance(basis, Number)])
    set_rp = set([basis for basis in parsed_input_data[f'{BANDS.rp}_basis_function_id'] if isinstance(basis, Number)])
    return remove_nans(set_bp).union(remove_nans(set_rp))


def get_design_matrices(sampling: np.ndarray, bases_config: pd.DataFrame) -> dict:
    """
    Get the design matrices corresponding to the input bases.

    Args:
        sampling (ndarray): 1D array containing the sampling grid.
        bases_config (NamedTuple): An object containing the configuration for all sets of basis functions.

    Returns:
        dict: The design matrices for the input list of bases.
    """
    bands_config = get_bands_config(bases_config)
    bp_config = bands_config.bpConfig
    rp_config = bands_config.rpConfig
    bp_config_dict = {field: getattr(bp_config, field) for field in bp_config._fields}
    rp_config_dict = {field: getattr(rp_config, field) for field in rp_config._fields}
    bands_config = {key: value for key, value in zip(BANDS, [bp_config_dict, rp_config_dict])}
    config_df = pd.DataFrame.from_dict(bands_config, orient='index')
    return {band: SampledBasisFunctions.from_config(sampling, config_df.loc[[band]]) for band in BANDS}



# flake8: noqa
from .calibrator.calibrator import calibrate
from .cholesky.cholesky import get_chi2, get_inverse_covariance_matrix, get_inverse_square_root_covariance_matrix
from .converter.converter import convert
from .core.dispersion_function import pwl_to_wl, wl_to_pwl, pwl_range, wl_range
from .core.version import __version__
from .error_correction.error_correction import apply_error_correction
from .generator.generator import generate
from .generator.photometric_system import PhotometricSystem, load_additional_systems, remove_additional_systems
from .plotter.plot_spectra import plot_spectra

__all__ = ['calibrate', 'get_chi2', 'get_inverse_covariance_matrix', 'get_inverse_square_root_covariance_matrix',
           'convert', 'pwl_to_wl', 'wl_to_pwl', 'pwl_range', 'wl_range', 'apply_error_correction', 'generate',
           'PhotometricSystem', 'load_additional_systems', 'remove_additional_systems', 'plot_spectra',
           '__version__']



"""
parse_internal_sampled.py
====================================
Module to parse input files containing internally calibrated sampled spectra.
"""

from .parse_generic import GenericParser


class InternalSampledParser(GenericParser):
    """
    Parser for internally calibrated sampled spectra.
    """

    def _parse_csv(self, csv_file):
        """
        Parse the input CSV file and store the result in a pandas DataFrame if it contains internally calibrated sampled
            spectra.

        Args:
            csv_file (str): Path to a CSV file.

        Returns:
            DataFrame: Pandas DataFrame representing the CSV file.
        """
        return super()._parse_csv(csv_file, _array_columns=['flux', 'error'])



from gaiaxpy.input_reader.file_reader import FileReader


class LocalFileReader(FileReader):

    def __init__(self, file_parser_selector, file, truncation, additional_columns=None, selector=None,
                 disable_info=False):
        super().__init__(file_parser_selector, file, truncation, additional_columns, selector, disable_info)



"""
absolute_sampled_spectrum.py
====================================
Module to represent an absolute sampled spectrum.
"""

from numbers import Number

import numpy as np

from .sampled_spectrum import SampledSpectrum
from .utils import _list_to_array
from ..core.custom_errors import NoBandsAvailableError
from ..core.generic_functions import correlation_from_covariance


class AbsoluteSampledSpectrum(SampledSpectrum):
    """
    A spectrum calibrated onto the absolute system of wavelength and flux. The spectrum is represented by a set of
    discrete measurements or samples.
    """

    def __init__(self, source_id, xp_spectra, sampled_bases, merge, truncation=None, with_correlation=False):
        """
        Initialise an absolute sampled spectrum.

        Args:
            source_id (str): Source identifier.
            xp_spectra (dict): A dictionary containing the BP and RP continuous spectra.
            sampled_bases (dict): The set of basis functions sampled onto the grid defining the resolution of the final
                sampled spectrum.
            merge (dict): The weighting factors for BP and RP sampled onto the grid defining the resolution of the final
                sampled spectrum.
            truncation (dict): Number of bases to be used for this spectrum. The set of bases functions used for the
                continuous representation of the spectra has been optimised to ensure that the first bases are the ones
                that contribute most. In many cases, the last bases contribution will be below the noise. Truncation of
                the basis function set to preserve only the significant bases is optional. By default, no truncation
                will be applied, i.e. all bases will be used.
            with_correlation (bool): Whether correlation information should be computed.
        """
        self.truncation = dict() if truncation is None else truncation
        self.available_bands = self.get_available_bands(xp_spectra)
        if not self.available_bands:
            raise NoBandsAvailableError()
        pos = sampled_bases[self.available_bands[0]].get_sampling_grid()
        SampledSpectrum.__init__(self, source_id, pos)
        self.pos = pos

    @staticmethod
    def get_available_bands(xp_spectra):
        def __is_available_band(_xp_spectra, band):
            return isinstance(_xp_spectra[band].covariance, np.ndarray)

        return [band for band in xp_spectra.keys() if __is_available_band(xp_spectra, band)]

    def generate_spectra(self, xp_spectra, sampled_bases, with_correlation):
        split_spectrum = {band: dict() for band in self.available_bands}
        for band in self.available_bands:
            band_truncation = self.truncation.get(band)
            split_spectrum[band]['xp_spectra'] = xp_spectra[band]
            stdev = split_spectrum[band]['xp_spectra'].get_standard_deviation()
            design_matrix = sampled_bases[band].get_design_matrix()
            spectra_covariance = split_spectrum[band]['xp_spectra'].get_covariance()
            coefficients = split_spectrum[band]['xp_spectra'].get_coefficients()
            if isinstance(band_truncation, Number) and band_truncation > 0:
                design_matrix = design_matrix[:band_truncation][:]
                spectra_covariance = spectra_covariance[:band_truncation, :band_truncation]
                coefficients = coefficients[:band_truncation]
            split_spectrum[band]['flux'] = self._sample_flux(coefficients, design_matrix)
            split_spectrum[band]['error'] = self._sample_error(spectra_covariance, design_matrix, stdev)
            if with_correlation:
                split_spectrum[band]['cov'] = self._sample_covariance(spectra_covariance, design_matrix)
                split_spectrum[band]['stdev'] = stdev
        return split_spectrum

    def __merge_output(self, split_spectrum, merge, with_correlation):
        raise NotImplementedError('Method not implemented for base class.')

    def _get_fluxes(self):
        return self.flux

    def _get_flux_errors(self):
        return self.error

    def get_positions(self):
        return self.pos

    @classmethod
    def get_units(cls):
        return {'flux': 'W.nm**-1.m**-2', 'flux_error': 'W.nm**-1.m**-2', 'pos': 'nm'}

    @classmethod
    def get_flux_label(cls):
        return r'Flux [$\mathrm{W\,nm^{-1}\,m^{-2}}$]'

    @classmethod
    def get_position_label(cls):
        return r'Wavelength [$\mathrm{nm}$]'

    def spectrum_to_dict(self, with_correlation):
        """
        Represent the spectrum as a dictionary.

        Returns:
            dict: A dictionary populated with the minimum set of parameters that need to be stored for this object.
                This is optimised for writing large number of sampled spectra and for this reason the array of positions
                is NOT included as it is expected to be the same for a batch of spectra. The array of positions can be
                retrieved calling the sampling_to_dict method.
        """
        spectrum_dict = {'source_id': self.source_id, 'flux': _list_to_array(self.flux),
                         'flux_error': _list_to_array(self.error)}
        if with_correlation:
            full_correlation = correlation_from_covariance(self.covariance)
            spectrum_dict['correlation'] = full_correlation[np.tril_indices(full_correlation.shape[0], k=-1)]
        return spectrum_dict

    def _sampling_to_dict(self):
        """
        Represent the sampling grid as a dictionary.

        Returns:
            dict: A dictionary populated with the sampling grid used for this spectrum.
        """
        return {'pos': _list_to_array(self.pos)}



from xml.etree import ElementTree

import numpy as np


def get_file_root(xml_file):
    return ElementTree.parse(xml_file).getroot()


def iterative_find(x_root, tag_list):
    def find_in_root(_x_root, _tag):
        return _x_root.find(_tag)

    result = find_in_root(x_root, tag_list[0])
    for tag in tag_list[1:]:
        result = find_in_root(result, tag)
    return result


def parse_array(x_root, tag):
    values, _ = get_array_text(x_root, tag)
    return np.array([float(element) for element in values])


def get_xp_merge(x_root):
    output = []
    tags = ['sampleMeanWavelengths', 'bpWeights', 'rpWeights']
    for tag in tags:
        xp_merge_array = iterative_find(x_root, ['XpMerge', tag])
        xp_merge_array = np.array([float(element.text) for element in xp_merge_array])
        output.append(xp_merge_array)
    return tuple(output)


def get_array_text(x_root, tag):
    result = x_root.find(tag)
    result = [element.text for element in result] if result is not None else None
    length = len(result) if result else None
    return result, length


def get_xp_sampling_matrix(x_root, xp, n_bands):
    xp_config = iterative_find(x_root, ['XpSampling', f'{xp.lower()}SampledBases'])
    xp_dimension = int(xp_config.attrib['dimension'])
    xp_sampling = np.array([float(element.text) for element in xp_config])
    if not n_bands:
        n_bands = len(xp_sampling) // xp_dimension
    xp_sampling = xp_sampling.reshape(n_bands, xp_dimension)
    return np.transpose(xp_sampling)



import numpy as np

from gaiaxpy.core.satellite import BANDS, RP_WL, BP_WL
from gaiaxpy.spectrum.absolute_sampled_spectrum import AbsoluteSampledSpectrum


class CalibrationAbsoluteSampledSpectrum(AbsoluteSampledSpectrum):

    def __init__(self, source_id, xp_spectra, sampled_bases, merge, truncation=None, with_correlation=False):
        super().__init__(source_id, xp_spectra, sampled_bases, merge, truncation=truncation,
                         with_correlation=with_correlation)
        split_spectrum = self.generate_spectra(xp_spectra, sampled_bases, with_correlation=with_correlation)
        self.__merge_output(split_spectrum, merge, with_correlation=with_correlation)

    def __merge_output(self, split_spectrum, merge, with_correlation):
        n_bands = len(split_spectrum.keys())
        # If both bands are present
        if n_bands == 2:
            self.flux = np.add(np.multiply(split_spectrum[BANDS.bp]['flux'], merge[BANDS.bp]),
                               np.multiply(split_spectrum[BANDS.rp]['flux'], merge[BANDS.rp]))
            self.error = np.sqrt(np.add(np.multiply(split_spectrum[BANDS.bp]['error'] ** 2, merge[BANDS.bp] ** 2),
                                        np.multiply(split_spectrum[BANDS.rp]['error'] ** 2, merge[BANDS.rp] ** 2)))
            if with_correlation:
                self.covariance = np.add(np.multiply(split_spectrum[BANDS.bp]['cov'], merge[BANDS.bp]),
                                         np.multiply(split_spectrum[BANDS.rp]['cov'], merge[BANDS.rp]))
        # If only one is
        elif n_bands == 1:
            existing_band, spectrum = list(split_spectrum.items())[0]
            self.flux = spectrum['flux']
            self.error = spectrum['error']
            if with_correlation:
                self.covariance = spectrum['cov']
            # Patch values if a band is missing
            masked_pos = self.pos.copy()
            masked_pos = masked_pos.astype(float)
            if existing_band == BANDS.rp:
                masked_pos[masked_pos <= RP_WL.low] = np.nan
            elif existing_band == BANDS.bp:
                masked_pos[masked_pos >= BP_WL.high] = np.nan
            else:
                raise ValueError(f'Band {existing_band} is not a valid band.')
            # Get the indices of all the values in pos that are smaller than the lowest RP range value
            self.flux[np.argwhere(np.isnan(masked_pos))] = np.nan
            self.error[np.argwhere(np.isnan(masked_pos))] = np.nan
            if with_correlation:
                self.covariance[:, np.argwhere(np.isnan(masked_pos))] = np.nan
                self.covariance[np.argwhere(np.isnan(masked_pos)), :] = np.nan



"""
input_validator.py
====================================
Module to validate the input data columns, etc.
"""
from gaiaxpy.core.generic_functions import _warning


def validate_required_columns(content_columns, requested_columns):
    """
    Validate whether all mandatory columns are present in the content_columns.

    Args:
        content_columns (list): List of columns in the content.
        requested_columns (list): List of requested columns.

    Raises:
        ValueError: If some required columns are not present.
    """

    def __are_required_columns_present(_content_columns, _mandatory_columns):
        return all(column in _content_columns for column in _mandatory_columns)

    def __get_missing_columns(_content_columns, _required_columns):
        missing_cols = [col for col in _required_columns if col not in _content_columns]
        missing_cols_str = ', '.join(missing_cols)
        return missing_cols_str

    if not requested_columns:
        return
    elif not __are_required_columns_present(content_columns, requested_columns):
        raise ValueError(f'Some required columns are not present. Please check your data. '
                         f'Missing columns are: {__get_missing_columns(content_columns, requested_columns)}.')


def validate_save_arguments(default_output_file, given_output_file, default_output_format, given_output_format,
                            save_file):
    """
    Validate file saving input arguments.

    Args:
        default_output_file: Default output file name as shown in the function from which this function is being called.
        given_output_file: Output file name provided by the user.
        default_output_format: Default value for output format.
        given_output_format: Output format provided by the user.
        save_file: Argument passed to save_file parameter.
    """
    if save_file and not isinstance(save_file, bool):
        raise ValueError("Parameter 'save_file' must contain a boolean value.")
    # If the user input a number different to the default value, but didn't set save_file to True
    if default_output_file != given_output_file and not save_file:
        _warning("The argument 'output_file' was provided, but 'save_file' is set to False. Set 'save_file' to True to "
                 "store the output of the function.")
    if default_output_format != given_output_format and not save_file:
        _warning("The argument 'output_format' was provided, but 'save_file' is set to False. Set 'save_file' to True "
                 "to store the output of the function.")


def check_column_overwrite(additional_columns, required_columns):
    common_names = []
    for key, value in additional_columns.items():
        if key in required_columns:
            if not isinstance(additional_columns[key], list):
                raise TypeError('Argument additional_columns seems to be malformed.')
            if len(additional_columns[key]) == 1 and additional_columns[key][0] != key:
                common_names.append(key)
            elif len(additional_columns[key]) > 1:
                common_names.append(key)
        if isinstance(value, list) and len(value) == 1 and value[0] in required_columns and key != value[0]:
            common_names.append(key)
    if common_names:
        raise ValueError('One or more elements in the additional columns input will be renamed to a column'
                         ' required by GaiaXPy to compute the output. This is not permitted. The offending'
                         f' elements are: {", ".join(common_names)}.')



"""
continuous_spectra_data.py
====================================
Module to represent continuous spectra data.
"""

import warnings
from os.path import join
from pathlib import Path

import numpy as np
import pandas as pd
from astropy.io import fits
from astropy.io.votable.tree import Field, Resource, Table, VOTableFile
from astropy.units import UnitsWarning
from fastavro import parse_schema, writer
from fastavro.validation import validate_many

from gaiaxpy.core.satellite import BANDS
from .output_data import OutputData
from .utils import _add_ecsv_header, _build_ecsv_header, _generate_fits_header, _load_header_dict

warnings.filterwarnings('ignore', category=UnitsWarning)


class ContinuousSpectraData(OutputData):

    def __init__(self, data):
        super().__init__(data, None)

    def _save_avro(self, output_path, output_file):
        """
        Save the output spectra in AVRO format.

        Args:
            output_path (str): Path where to save the file.
            output_file (str): Name of the output file.
        """

        def _generate_avro_schema(_spectra_dicts):
            """
            Generate the AVRO schema required to store the output.

            Args:
                _spectra_dicts (list): A list of dictionaries containing spectra.

            Returns:
                dict: A dictionary containing the parsed schema that matches the input.
                list of dicts: A list of dictionaries with the modified input spectra
                according to the valid AVRO types.
            """
            field_to_type = {
                'source_id': 'long',
                'bp_standard_deviation': 'float', 'rp_standard_deviation': 'float',
                'bp_coefficients': 'string', 'rp_coefficients': 'string',
                'bp_coefficient_correlations': 'string', 'rp_coefficient_correlations': 'string',
                'bp_coefficient_errors': 'string', 'rp_coefficient_errors': 'string',
                'bp_n_parameters': 'int', 'rp_n_parameters': 'int',
                'bp_basis_function_id': 'int', 'rp_basis_function_id': 'int'
            }

            def build_field(keys):
                return [{'name': key, 'type': field_to_type[key]} for key in keys]

            schema = {'doc': 'Spectrum output.', 'name': 'Spectra', 'namespace': 'spectrum', 'type': 'record',
                      'fields': build_field(_spectra_dicts[0].keys())}
            # Spectrum fields to string
            for spectrum in _spectra_dicts:
                for field, _type in field_to_type.items():
                    if _type == 'string':
                        spectrum[field] = str(tuple(spectrum[field]))
            # Validate that records match the schema
            validate_many(_spectra_dicts, schema)
            return parse_schema(schema), _spectra_dicts

        data = self.data
        # List with one dictionary per source
        spectra_dicts = data.to_dict('records')
        parsed_schema, spectra_dicts = _generate_avro_schema(spectra_dicts)
        Path(output_path).mkdir(parents=True, exist_ok=True)
        output_path = join(output_path, f'{output_file}.avro')
        with open(output_path, 'wb') as output:
            writer(output, parsed_schema, spectra_dicts)

    def _save_csv(self, output_path, output_file):
        """
        Save the output spectra in CSV format.

        Args:
            output_path (str): Path where to save the file.
            output_file (str): Name of the output file.
        """
        spectra_df = self.data
        array_columns = [column for column in spectra_df.columns if isinstance(spectra_df[column].iloc[0], np.ndarray)]
        spectra_df[array_columns] = spectra_df[array_columns].apply(lambda col: col.apply(tuple)).astype('str')
        Path(output_path).mkdir(parents=True, exist_ok=True)
        output_path = join(output_path, f'{output_file}.csv')
        spectra_df.to_csv(output_path, index=False)

    def _save_ecsv(self, output_path, output_file):
        """
        Save the output spectra in ECSV format.

        Args:
            output_path (str): Path where to save the file.
            output_file (str): Name of the output file.
        """
        spectra_df = self.data
        array_columns = [column for column in spectra_df.columns if isinstance(spectra_df[column].iloc[0], np.ndarray)]
        header_lines = _build_ecsv_header(spectra_df)
        spectra_df[array_columns] = spectra_df[array_columns].apply(lambda col: col.apply(tuple)).astype('str')
        Path(output_path).mkdir(parents=True, exist_ok=True)
        spectra_df.to_csv(join(output_path, f'{output_file}.ecsv'), index=False)
        _add_ecsv_header(header_lines, output_path, output_file)

    def _save_fits(self, output_path, output_file):
        """
        Save the output data in FITS format.

        Args:
            output_path (str): Path where to save the file.
            output_file (str): Name of the output file.
        """
        data = self.data
        coefficients_format = f"{len(data['bp_coefficients'].iloc[0])}D"  # D: double precision float
        correlations_format = f"{len(data['bp_coefficient_correlations'].iloc[0])}D"
        errors_format = f"{len(data['bp_coefficient_errors'].iloc[0])}E"  # E: single precision float
        # Define formats for each type according to FITS
        column_formats = {
            'source_id': 'K',
            'bp_standard_deviation': 'D', 'rp_standard_deviation': 'D',
            'bp_coefficients': coefficients_format, 'rp_coefficients': coefficients_format,
            'bp_coefficient_correlations': correlations_format, 'rp_coefficient_correlations': correlations_format,
            'bp_coefficient_errors': errors_format, 'rp_coefficient_errors': errors_format,
            'bp_n_parameters': 'I', 'rp_n_parameters': 'I',
            'bp_basis_function_id': 'I', 'rp_basis_function_id': 'I'}
        # create a list of HDUs
        hdu_list = list()
        # create a header to include the sampling
        hdr = fits.Header()
        primary_hdu = fits.PrimaryHDU(header=hdr)
        hdu_list.append(primary_hdu)
        # Create a dictionary to hold all the data
        output_by_column_dict = data.reset_index().to_dict(orient='list')
        # Remove index from output dict
        output_by_column_dict.pop('index', None)
        spectra_keys = output_by_column_dict.keys()
        data_type = data.attrs['data_type']
        units_dict = data_type.get_units()
        columns = [fits.Column(name=key, array=np.array(output_by_column_dict[key]), format=column_formats[key],
                               unit=units_dict.get(key, '')) for key in spectra_keys]
        header = _generate_fits_header(data, column_formats)
        hdu = fits.BinTableHDU.from_columns(columns, header=header)
        hdu_list.append(hdu)
        # Put all HDUs together
        hdul = fits.HDUList(hdu_list)
        # Write the file and replace it if it already exists
        Path(output_path).mkdir(parents=True, exist_ok=True)
        output_path = join(output_path, f'{output_file}.fits')
        hdul.writeto(output_path, overwrite=True)

    def _save_xml(self, output_path, output_file):
        """
        Save the output spectra in XML/VOTABLE format.

        Args:
            output_path (str): Path where to save the file.
            output_file (str): Name of the output file.
        """

        def _create_fields(_votable, _spectra_df):
            fields_datatypes = {'source_id': 'long',
                                f'{BANDS.bp}_standard_deviation': 'double', f'{BANDS.rp}_standard_deviation': 'double',
                                f'{BANDS.bp}_coefficients': 'double', f'{BANDS.rp}_coefficients': 'double',
                                f'{BANDS.bp}_coefficient_correlations': 'double',
                                f'{BANDS.rp}_coefficient_correlations': 'double',
                                f'{BANDS.bp}_coefficient_errors': 'float', f'{BANDS.rp}_coefficient_errors': 'float',
                                f'{BANDS.bp}_n_parameters': 'int', f'{BANDS.rp}_n_parameters': 'int',
                                f'{BANDS.bp}_basis_function_id': 'int', f'{BANDS.rp}_basis_function_id': 'int'}
            fields_array_size = {'source_id': '',
                                 f'{BANDS.bp}_standard_deviation': '', f'{BANDS.rp}_standard_deviation': '',
                                 f'{BANDS.bp}_coefficients': '*', f'{BANDS.rp}_coefficients': '*',
                                 f'{BANDS.bp}_coefficient_correlations': '*',
                                 f'{BANDS.rp}_coefficient_correlations': '*',
                                 f'{BANDS.bp}_coefficient_errors': '*', f'{BANDS.rp}_coefficient_errors': '*',
                                 f'{BANDS.bp}_n_parameters': '', f'{BANDS.rp}_n_parameters': '',
                                 f'{BANDS.bp}_basis_function_id': '', f'{BANDS.rp}_basis_function_id': ''}
            columns = _spectra_df.columns
            data_type = _spectra_df.attrs['data_type']
            units_dict = data_type.get_units()
            header_dict = _load_header_dict()
            _fields = [Field(_votable, name=column, datatype=fields_datatypes[column],
                             arraysize=fields_array_size[column], ucd=header_dict.get(column, dict()).get('meta', ''),
                             unit=units_dict.get(column, '')) if fields_array_size[column] != '' else
                       Field(_votable, name=column, datatype=fields_datatypes[column],
                             ucd=header_dict.get(column, dict()).get('meta', ''), unit=units_dict.get(column, ''))
                       for column in columns]
            for _field in _fields:
                _field.description = header_dict.get(_field.name, dict()).get('description', '')
            return _fields

        spectra_df = self.data
        # Create a new VOTable file
        votable = VOTableFile()
        # Add a resource
        resource = Resource()
        votable.resources.append(resource)
        # Add a table for the spectra (and add the sampling as metadata)
        spectra_table = Table(votable)
        resource.tables.append(spectra_table)
        # Add spectrum fields
        fields = _create_fields(votable, spectra_df)
        spectra_table.fields.extend(fields)
        # Create the record arrays, with the given number of rows
        spectra_table.create_arrays(len(spectra_df))
        for index, row in enumerate(spectra_df.to_dict('records')):
            spectra_table.array[index] = tuple([row[column] for column in spectra_df.columns])
        # Write to a file
        Path(output_path).mkdir(parents=True, exist_ok=True)
        output_path = join(output_path, f'{output_file}.xml')
        votable.to_xml(output_path)

    def _get_spectra_df(self):
        data = self.data
        spectra_bp_df = pd.DataFrame.from_records([spectrum[BANDS.bp].spectrum_to_dict() for spectrum in data])
        spectra_rp_df = pd.DataFrame.from_records([spectrum[BANDS.rp].spectrum_to_dict() for spectrum in data])
        spectra_df = spectra_bp_df.merge(spectra_rp_df, on='source_id', how='outer')
        for col in spectra_df.columns:
            if 'xp' in col:
                spectra_df = spectra_df.drop(col, axis=1)
            else:
                if '_x' in col:
                    col_new = col.replace('_x', '')
                    col_new = f'{BANDS.bp}_' + col_new
                    spectra_df = spectra_df.rename(columns={col: col_new})
                if '_y' in col:
                    col_new = col.replace('_y', '')
                    col_new = f'{BANDS.rp}_' + col_new
                    spectra_df = spectra_df.rename(columns={col: col_new})
        return spectra_df



import numpy as np
import pandas as pd

from gaiaxpy.core.generic_functions import str_to_array

# Avoid warning, false positive
pd.options.mode.chained_assignment = None


class DataFrameStringArrayReader(object):

    def __init__(self, content, array_columns):
        self.content = content.copy()
        self.array_columns = array_columns

    def _parse_parenthesis_arrays(self):
        df = self.content
        array_columns = self.array_columns
        for column in array_columns:
            df[column] = df[column].apply(lambda x: str_to_array(x))
        return df

    def _parse_brackets_arrays(self):
        df = self.content
        array_columns = self.array_columns
        for column in array_columns:
            df[column] = df[column].map(np.array)
        return df

    def read(self):
        def __get_enclosing_element(_df, _array_columns):
            for row in _df.to_dict('records'):
                for column in _array_columns:
                    if isinstance(row[column], str):
                        if row[column][0] in ['(', '[']:
                            return row[column][0]
                    else:
                        continue

        df = self.content
        array_columns = self.array_columns
        enclosing = __get_enclosing_element(df, array_columns)
        if enclosing == '(':
            df = self._parse_parenthesis_arrays()
        elif enclosing == '[':
            df = self._parse_brackets_arrays()
        return df



__CAL_MANDATORY_COLS = ['source_id', 'bp_n_parameters', 'bp_standard_deviation', 'rp_n_parameters',
                        'rp_standard_deviation']
__CON_MANDATORY_COLS = ['source_id', 'bp_n_parameters', 'bp_standard_deviation', 'rp_n_parameters',
                        'rp_standard_deviation']
__GEN_MANDATORY_COLS = ['source_id', 'bp_n_parameters', 'bp_standard_deviation', 'rp_n_parameters',
                        'rp_standard_deviation']
__INV_MANDATORY_COLS = ['source_id', 'bp_n_parameters', 'rp_n_parameters', 'bp_standard_deviation',
                        'rp_standard_deviation']

COV_INPUT_COLUMNS = ['bp_coefficient_covariances', 'rp_coefficient_covariances']
CORR_INPUT_COLUMNS = ['bp_coefficients', 'bp_coefficient_errors', 'bp_coefficient_correlations',
                      'rp_coefficients', 'rp_coefficient_errors', 'rp_coefficient_correlations']

MANDATORY_INPUT_COLS = {'_apply_colour_equation': __GEN_MANDATORY_COLS, '_apply_error_correction': __GEN_MANDATORY_COLS,
                        '_calibrate': __CAL_MANDATORY_COLS, 'calibrate': __CAL_MANDATORY_COLS,
                        '_convert': __CON_MANDATORY_COLS, 'convert': __CON_MANDATORY_COLS,
                        '_generate': __GEN_MANDATORY_COLS, 'generate': __GEN_MANDATORY_COLS,
                        'get_inverse_covariance_matrix': __INV_MANDATORY_COLS,
                        'get_inverse_square_root_covariance_matrix': __INV_MANDATORY_COLS}

TRUNCATION_COLS = ['bp_n_relevant_bases', 'rp_n_relevant_bases']

# Output columns
CAL_CORR_OUTPUT_COLS = ['correlation']
__CAL_OUTPUT_COLS = ['source_id', 'flux', 'flux_error']

CON_CORR_OUTPUT_COLS = ['correlation', 'standard_deviation']
__CON_OUTPUT_COLS = ['source_id', 'xp', 'flux', 'flux_error']

# Generator columns are not fixed but variable, depending on the systems involved; no correlation columns either.

# Whether all columns are present depends on the band(s) being requested
__INV_OUTPUT_COLS = ['source_id', 'bp_inverse_covariance', 'rp_inverse_covariance']
__INV_SQRT_OUTPUT_COLS = ['source_id', 'bp_inverse_square_root_covariance_matrix',
                          'rp_inverse_square_root_covariance_matrix']

# apply_error_correction should output the same columns that were received in the input, and will not accept the
# additional_columns argument



from pathlib import Path
from typing import Union, Optional

import pandas as pd

from gaiaxpy.colour_equation.xp_filter_system_colour_equation import _apply_colour_equation
from gaiaxpy.core.generic_functions import cast_output, format_additional_columns, validate_photometric_system
from gaiaxpy.error_correction.error_correction import _apply_error_correction
from gaiaxpy.input_reader.input_reader import InputReader
from gaiaxpy.output.photometry_data import PhotometryData
from .multi_synthetic_photometry_generator import MultiSyntheticPhotometryGenerator
from .photometric_system import PhotometricSystem
from ..core.input_validator import validate_save_arguments
from ..file_parser.cast import _cast


def generate(input_object: Union[list, Path, pd.DataFrame, str], photometric_system: Union[list, PhotometricSystem],
             truncation: bool = False, output_path: Union[Path, str] = '.',
             output_file: str = 'output_synthetic_photometry',
             output_format: str = None, save_file: bool = True, error_correction: bool = False,
             additional_columns: Optional[Union[dict, list, str]] = None, username: str = None, password: str = None) \
        -> pd.DataFrame:
    """
    Synthetic photometry utility: generates synthetic photometry in a set of available systems from the input
    internally-calibrated continuously-represented mean spectra.
    Some standardised photometric systems include a colour-correction to the U bands which will be applied
    automatically when generating the corresponding synthetic photometry.

    Args:
        input_object (list/Path/pd.DataFrame/str): Path to the file containing the mean spectra as downloaded from
            the archive in their continuous representation, a list of sources ids (string or long), or a pandas
            DataFrame.
        photometric_system (list/PhotometricSystem): Desired photometric system or list of photometric systems.
        truncation (bool): Toggle truncation of the set of bases. The level of truncation to be applied is defined
            by the recommended value in the input files.
        output_path (Path/str): Path where to save the output data.
        output_file (str): Name of the output file without extension (e.g. 'my_file').
        output_format (str): Desired output format. If no format is given, the output file format will be the same as
            the input file (e.g. 'csv').
        save_file (bool): Whether to save the output in a file or not. If false, output_format and output_file_name will
            be ignored.
        error_correction (bool): Whether to apply to the photometric errors the tabulated factors to mitigate
            underestimated errors (see Montegriffo et al., 2022, for more details).
        additional_columns (str/list): List of additional columns to include in the output. The columns must be requested
            columns must be available in the input (files, DataFrames) or in the Archive response (lists, queries).
        username (str): Cosmos username, only suggested when input_object is a list or ADQL query.
        password (str): Cosmos password, only suggested when input_object is a list or ADQL query.

    Returns:
        DataFrame: A DataFrame of all synthetic photometry results.
    """
    return _generate(input_object=input_object, photometric_system=photometric_system, truncation=truncation,
                     output_path=output_path, output_file=output_file, output_format=output_format,
                     save_file=save_file, error_correction=error_correction, additional_columns=additional_columns,
                     username=username, password=password)


def _generate(input_object: Union[list, Path, pd.DataFrame, str], *, photometric_system: Union[list, PhotometricSystem],
              truncation: bool = False, output_path: Union[Path, str] = '.',
              output_file: str = 'output_synthetic_photometry', output_format: str = None, save_file: bool = True,
              error_correction: bool = False, additional_columns: Optional[Union[dict, list, str]] = None,
              selector=None, username: str = None, password: str = None, bp_model: str = 'v375wi',
              rp_model: str = 'v142r') -> pd.DataFrame:
    """
    Internal function of the calibration utility. Refer to "generate".

    Args:
        selector (function): Function to filter AVRO records. The records returned will be the
            ones for which the function returns True. The field names used in the selector
            function should match the ones in the AVRO schema as the filter is run before any
            column renaming happens. If selector is not None and the input is not an AVRO file,
            a SelectorNotImplementedError will be raised.
        bp_model (str): The bp model.
        rp_model (str): The rp model.
    """

    def __is_gaia_initially_in_systems(_internal_photometric_system: list,
                                       _gaia_system: PhotometricSystem = PhotometricSystem.Gaia_DR3_Vega):
        """
        Check whether Gaia DR3 is originally in the input photometric systems.

        Args:
            _internal_photometric_system (list): List of photometric systems.
            _gaia_system (PhotometricSystem): Gaia DR3 system.

        Returns:
            bool: True if Gaia DR3 is in the list, False otherwise.
        """
        gaia_system_name = _gaia_system.get_system_name()
        return any([item.get_system_name() == gaia_system_name for item in _internal_photometric_system])

    validate_photometric_system(photometric_system)
    validate_save_arguments(generate.__defaults__[2], output_file, generate.__defaults__[3], output_format, save_file)
    # Prepare systems, keep track of original systems (especially required for error_correction)
    internal_phot_system = photometric_system.copy() if isinstance(photometric_system, list) else (
        [photometric_system].copy())
    gaia_system = PhotometricSystem.Gaia_DR3_Vega
    is_gaia_in_input = __is_gaia_initially_in_systems(internal_phot_system)
    if error_correction and not is_gaia_in_input:
        internal_phot_system.append(gaia_system)
    additional_columns = format_additional_columns(additional_columns)
    # Read input data
    parsed_input_data, extension = InputReader(input_object, generate, truncation=truncation,
                                               additional_columns=additional_columns, selector=selector, user=username,
                                               password=password).read()
    additional_data = parsed_input_data[list(additional_columns.keys())]
    # Generate photometry
    phot_generator = MultiSyntheticPhotometryGenerator(internal_phot_system, bp_model=bp_model, rp_model=rp_model)
    photometry_df = phot_generator.generate(parsed_input_data, extension, output_file=None, output_format=None,
                                            save_file=False, truncation=truncation)
    photometry_df = _apply_colour_equation(photometry_df, photometric_system=internal_phot_system, save_file=False,
                                           disable_info=True)
    if error_correction:
        photometry_df = _apply_error_correction(photometry_df, photometric_system=photometric_system, save_file=False,
                                                disable_info=True)
        if not is_gaia_in_input:  # Remove Gaia_DR3_Vega system from the final result
            gaia_label = gaia_system.get_system_label()
            gaia_columns = [column for column in photometry_df if column.startswith(gaia_label)]
            photometry_df = photometry_df.drop(columns=gaia_columns)
    additional_data = additional_data[[c for c in additional_data.columns if c not in photometry_df.columns]]
    photometry_df = pd.concat([photometry_df, additional_data], axis=1)
    photometry_df = cast_output(photometry_df)
    # Save data
    output_data = PhotometryData(photometry_df)
    output_data.save(save_file, output_path, output_file, output_format, extension)
    return _cast(photometry_df)



from ast import literal_eval
from os.path import abspath, dirname, join

import numpy as np
import pandas as pd
from astropy.io import fits
from numpy import ndarray


def pandas_from_records(lst):
    return pd.DataFrame.from_records([record.to_dict_func for record in lst])


def _array_to_standard(array, extension='csv'):
    """
    Converts an array to a tuple so that its string representation corresponds to the archive standard where a list is
        represented using parentheses and commas, i.e.: "(elem1, elem2)".

    Args:
        array (ndarray): An array of floats.
        extension (str): Format to use 'csv' means use round brackets, 'ecsv' means use square brackets.

    Returns:
        tuple: The array converted to a tuple.
    """
    if array is None:
        return array
    if not isinstance(array, ndarray):
        raise ValueError('Input must be a NumPy array.')

    def convert_ecsv(row):
        # Square brackets should be used and nan values should be shown as null (no quotes)
        if np.isnan(np.min(row)):
            return '[' + ', '.join(['null' if np.isnan(value) else str(value) for value in row]) + ']'
        return list(row)

    conversion_functions = {'csv': tuple, 'ecsv': convert_ecsv}
    if array.ndim > 1:
        conversion_function = conversion_functions[extension]
        return conversion_function([conversion_function(row) for row in array])
    else:
        conversion_function = conversion_functions[extension]
        return conversion_function(array)


def _get_array_columns(df):
    return [column for column in df.columns if isinstance(df[column].iloc[0], ndarray)]


def _get_sampling_dict(positions):
    return {'pos': _array_to_standard(positions)}


def _load_header_dict():
    current_path = dirname(abspath(__file__))
    header_dictionary_path = join(current_path, 'ecsv_headers', 'headers_dict.txt')
    with open(header_dictionary_path) as f:
        data = f.read()
    # Load header dictionary
    header_dict = literal_eval(data)
    return header_dict


def _get_col_subtype_len(_df, _column):
    for index in range(0, len(_df)):
        try:
            return len(_df[_column].iloc[index])
        except TypeError:
            pass
    raise ValueError('All arrays in the data seem to be empty. This should never happen.')


def _build_ecsv_header(df, positions=None):
    positions = None if positions is None else str(list(positions))
    columns = df.columns
    header_dict = _load_header_dict()
    header = _initialise_header()
    data_type = df.attrs['data_type']
    units_dict = data_type.get_units()
    for column in columns:
        current_column = header_dict[column]
        header.append('# -')
        header.append(f'#   name: {column}')
        header.append(f'#   datatype: {current_column["datatype"]}')
        if 'subtype' in current_column.keys():
            header.append(
                f'#   subtype: {current_column["subtype"].replace("null", str(_get_col_subtype_len(df, column)))}')
        header.append(f'#   description: {current_column["description"]}')
        if units_dict.get(column, None):
            header.append(f'#   unit: {units_dict[column]}')
        if current_column.get('meta', None):
            header.append('#   meta:')
            header.append(f'#     ucd: {current_column["meta"]}')
    if positions:
        header.append('# meta:')
        header.append(f'#   sampling: {positions}')
    return '\n'.join(header) + '\n'


def _initialise_header():
    return ["# %ECSV 1.0", "# ---", "# delimiter: ','", "# datatype:"]


def _build_photometry_header(columns):
    header_dict = _load_header_dict()
    header = _initialise_header()
    for column in columns:
        header.append('# -')
        header.append(f'#   name: {column}')
        if column != 'source_id':
            if '_flux_error_' in column:
                parameter = '_flux_error_'
            elif '_flux_' in column:
                parameter = '_flux_'
            elif '_mag_' in column:
                parameter = '_mag_'
            system, band = column.split(parameter)
            parameter = f'phot{parameter}'[:-1]
            header.append(f'#   datatype: {header_dict[parameter]["datatype"]}')
            header.append(f'#   description: {header_dict[parameter]["description"]} {band} band')
        else:
            header.append(f'#   datatype: {header_dict[column]["datatype"]}')
            header.append(f'#   description: {header_dict[column]["description"]}')
    return '\n'.join(header) + '\n'


def _add_ecsv_header(header, output_path, output_file):
    with open(join(output_path, f'{output_file}.ecsv'), "r+") as f:
        s = f.read()
        f.seek(0)
        f.write(header + s)


def _generate_fits_header(_data, _column_formats):
    data_type = _data.attrs['data_type']
    units_dict = data_type.get_units()
    header_dict = _load_header_dict()
    cards = list()
    for index, column in enumerate(_data.columns):
        cards.append((f'TTYPE{index + 1}', column))
        cards.append((f'TFORM{index + 1}', _column_formats.get(column, '')))
        cards.append((f'TCOMM{index + 1}', header_dict.get(column, dict()).get('description', '')))
        cards.append((f'TUCD{index + 1}', header_dict.get(column, dict()).get('meta', '')))
        cards.append((f'TUNIT{index + 1}', units_dict.get(column, '')))
    header = fits.Header(cards=cards)
    return header


def _build_line_header(columns):
    header_dict = _load_header_dict()
    header = _initialise_header()
    for column in columns:
        header.append('# -')
        header.append(f'#   name: {column}')
        header.append(f'#   datatype: {header_dict[column]["datatype"]}')
        header.append(f'#   description: {header_dict[column]["description"]}')
    return '\n'.join(header) + '\n'



"""
generic_spectrum.py
====================================
Module to represent a generic spectrum.
"""
from typing import Union


class Spectrum(object):
    """
    Base spectrum. Contain only the source ID.
    """

    def __init__(self, source_id: Union[str, int]):
        """
        Initialise a spectrum.

        Args:
            source_id (str): Source identifier. Can be a Gaia source ID (long) or any other source identifier.
        """
        self.source_id = source_id

    def get_source_id(self):
        """
        Get the source ID of the spectrum.

        Returns:
            str: Source identifier.
        """
        return self.source_id



"""
synthetic_photometry_generator.py
====================================
Module for the generation of synthetic photometry.
"""

from configparser import ConfigParser

from gaiaxpy.config.paths import config_ini_file
from gaiaxpy.core.satellite import BANDS
from gaiaxpy.spectrum.sampled_basis_functions import SampledBasisFunctions
from gaiaxpy.spectrum.single_synthetic_photometry import SingleSyntheticPhotometry
from gaiaxpy.spectrum.utils import get_covariance_matrix
from gaiaxpy.spectrum.xp_continuous_spectrum import XpContinuousSpectrum

config_parser = ConfigParser()
config_parser.read(config_ini_file)


class SyntheticPhotometryGenerator(object):
    def generate(self, parsed_input_data, extension, output_file, output_format, save_file, truncation):
        raise ValueError('Method not defined for base class.')

    def _get_sampled_basis_functions(self, xp_sampling, xp_sampling_grid):
        return {band: SampledBasisFunctions.from_design_matrix(xp_sampling_grid, xp_sampling[band]) for band in BANDS}

    def _create_photometry_list(self, parsed_input_data, photometric_system, sampled_basis_func, truncation, xp_merge):
        parsed_input_data_dict = parsed_input_data.to_dict('records')
        return (_generate_synthetic_photometry(row, sampled_basis_func, xp_merge, truncation, photometric_system) for
                row in parsed_input_data_dict)


def _generate_synthetic_photometry(row, design_matrix, merge, truncation, photometric_system):
    """
    Create the synthetic photometry from the input continuously-represented mean spectrum and design matrix.

    Args:
        row (DataFrame): Single row in a DataFrame containing the entry for one source in the mean spectra file. This
            will include columns for both bands (although one could be missing).
        design_matrix (ndarray): 2D array containing the basis functions sampled for the specific photometric system.
        merge (dict): Dictionary containing an array of weights per BP and one for RP. These have one value per sample
            and define the contributions from BP and RP to the joined absolute spectrum.
        truncation (bool): Toggle truncation of the set of bases.
        photometric_system (obj): Photometric system object containing the zero-points.

    Returns:
        SingleSyntheticPhotometry: The output synthetic photometry.
    """
    cont_dict = {band: XpContinuousSpectrum(row['source_id'], band.upper(), row[f'{band}_coefficients'],
                                            get_covariance_matrix(row, band), row[f'{band}_standard_deviation'])
                 for band in BANDS}
    truncation = {band: row[f'{band}_n_relevant_bases'] for band in BANDS} if truncation else None
    return SingleSyntheticPhotometry(row['source_id'], cont_dict, design_matrix, merge, truncation, photometric_system)



def execute_answer(action, message=None, args=None):
    """
    Executes the given action with provided arguments. Prints the given message if provided.
    """
    if action:
        if args:
            action(args)
        else:
            action()
        if message:
            print(message)


def get_yes_no_answer(question, yes_action, no_action, yes_message=None, no_message=None, yes_args=None, no_args=None):
    """
    Prompts user for a yes/no answer to the given question.
    Calls yes_action with yes_args if the answer is yes, and no_action with no_args if the answer is no.
    Prints yes_message after executing yes_action, and no_message after executing no_action.
    """
    default = 'y'
    yes_choices = ['yes', default]
    no_choices = ['no', 'n']
    while True:
        user_input = input(question) or default
        if user_input.lower() in yes_choices:
            execute_answer(yes_action, yes_message, yes_args)
            break
        elif user_input.lower() in no_choices:
            execute_answer(no_action, no_message, no_args)
            break
        else:
            print('Please type yes or no.')
            continue



"""
sampled_spectrum.py
====================================
Module to represent a sampled spectrum.
"""

import numpy as np
from numpy import ndarray, nan

from .generic_spectrum import Spectrum


class SampledSpectrum(Spectrum):
    """
    A spectrum defined by a set of discrete measurements. Each measurement is defined by a position in wavelength (or
    pseudo-wavelength), a measured flux and an associated flux error.
    Specific implementations of this class will define the units in use for positions and fluxes.
    """

    def __init__(self, source_id, sampling_grid):
        """
        Initialise a sampled spectrum.

        Args:
            source_id (str): Source identifier.
            sampling_grid (ndarray): 1D array containing the positions of the samples (or discrete measurements) for
                this spectrum.
        """
        Spectrum.__init__(self, source_id)
        self.n_samples = np.size(sampling_grid)

        self.pos = None
        self.flux = None
        self.error = None
        self.covariance = None

    def _get_fluxes(self):
        """
        Get the flux samples.

        Returns:
            ndarray: 1D array containing all flux samples.
        """
        return self.flux

    def _get_flux_errors(self):
        """
        Get the flux errors of each sample.

        Returns:
            ndarray: 1D array containing the error in flux for all samples.
        """
        return self.error

    def _get_covariance(self):
        """
        Get the flux covariance matrix.

        Returns:
            ndarray: 2D array containing the flux covariance matrix.
        """
        return self.covariance

    def get_positions(self):
        """
        Get the positions of all samples.

        Returns:
            ndarray: 1D array containing the position of all samples.
        """
        return self.pos

    def get_flux_label(self):
        """
        Get the labels describing the flux measurements.

        Returns:
            str: Short description of the flux measurements, including the units.
        """
        return ""

    def get_position_label(self):
        """
        Get the positions of the samples, including the units.

        Returns:
            str: Short description of the positions of the samples.
        """
        return ""

    def _get_inputs(self, spectrum):
        return spectrum.get_positions(), spectrum._get_fluxes(), spectrum._get_flux_errors()

    @staticmethod
    def _sample_flux(coefficients, design_matrix):
        """
        Given a set of coefficients to be applied to a set of basis functions and a design matrix containing the
        evaluation of each basis function at the positions corresponding to the samples, this method computes the flux
        values for each sample.

        Args:
            coefficients (ndarray): 1D array containing the coefficients multiplying the basis functions in the
                continuous representation.
            design_matrix (ndarray): 2D array containing the evaluation of the basis functions on the desired sampling
                grid.

        Returns:
            ndarray: 1D array containing the flux values for all samples.
        """
        if coefficients.shape[0] != 0 and coefficients.shape[0] != design_matrix.shape[0]:
            raise ValueError("Coefficients length doesn't match the design matrix dimension. Please make sure you're "
                             "using the correct input files and configuration.")
        if isinstance(coefficients, ndarray) and coefficients.size > 0:
            return coefficients @ design_matrix
        return nan

    @staticmethod
    def _sample_error(covariance, design_matrix, standard_deviation):
        """
        Given the covariance matrix and standard deviation of the least squares solution defining the continuous
        representation of the spectrum in terms of basis functions and a design matrix containing the evaluation of each
        basis function at the positions corresponding to the samples, this method computes the error associated to the
        flux value for each sample.

        Args:
            covariance (ndarray): 2D array containing the elements of the covariance matrix.
            design_matrix (ndarray): 2D array containing the evaluation of the basis functions on the desired sampling
                grid.
            standard_deviation (float): Standard deviation.

        Returns:
            ndarray: 1D array containing the errors in flux for all samples.
        """
        if isinstance(covariance, ndarray):
            return np.sqrt(
                np.sum(np.multiply(design_matrix.T @ covariance, design_matrix.T), axis=1)) * standard_deviation
        elif np.isnan(covariance):
            return float('NaN')
        else:
            raise TypeError('Covariance must be either a NumPy array or nan.')

    @staticmethod
    def _sample_covariance(covariance, design_matrix):
        """
        Given the covariance matrix and a design matrix containing the evaluation of each basis function at the
        positions corresponding to the samples, this method computes the covariance matrix of the sampled spectrum.

        Args:
            covariance (ndarray): 2D array containing the elements of the covariance matrix of the continuous
                representation.
            design_matrix (ndarray): 2D array containing the evaluation of the basis functions on the desired sampling
            grid.

        Returns:
            ndarray: 2D array containing the covariance matrix of the sampled spectrum.
        """
        return design_matrix.T @ covariance @ design_matrix



"""
parse_external.py
====================================
Module to parse input files containing externally calibrated sampled spectra.
"""

from .parse_generic import GenericParser

# Columns that contain arrays (as strings)
array_columns = ['wl', 'flux', 'flux_error']


class ExternalParser(GenericParser):
    """
    Parser for externally calibrated sampled spectra.
    """

    def __init__(self, requested_columns=None, additional_columns=None, selector=None, **kwargs):
        super().__init__()
        self.additional_columns = dict() if additional_columns is None else additional_columns
        self.requested_columns = requested_columns
        self.selector = selector
        if kwargs:
            self.address = kwargs.get('address', None)
            self.port = kwargs.get('port', None)

    def _parse_csv(self, csv_file, _array_columns=None, _matrix_columns=None):
        """
        Parse the input CSV file and store the result in a pandas DataFrame if it contains externally calibrated sampled
            spectra.

        Args:
            csv_file (str): Path to a CSV file.
            _array_columns (list): List of columns in the file that contain arrays as strings.
            _matrix_columns (list): Parameter required in the parser hierarchy. Not used in this function.

        Returns:
            DataFrame: Pandas DataFrame representing the CSV file.
        """
        if _array_columns is None:
            _array_columns = array_columns
        return super()._parse_csv(csv_file, _array_columns)

    def _parse_fits(self, fits_file, _array_columns=None, _matrix_columns=None):
        """
        Parse the input FITS file and store the result in a pandas DataFrame if it contains externally calibrated
            sampled spectra.

        Args:
            fits_file (str): Path to an FITS file.
            _array_columns (list): List of columns in the file that contain arrays as strings.

        Returns:
            DataFrame: Pandas DataFrame representing the FITS file.
        """
        if _array_columns is None:
            _array_columns = array_columns
        return super()._parse_fits(fits_file, _array_columns=_array_columns)

    def _parse_xml(self, xml_file, _array_columns=None, _matrix_columns=None):
        """
        Parse the input XML file and store the result in a pandas DataFrame if it contains externally calibrated sampled
            spectra.

        Args:
            xml_file (str): Path to an XML file.
            _array_columns (list): List of columns in the file that contain arrays as strings.

        Returns:
            DataFrame: Pandas DataFrame representing the XML file.
        """
        if _array_columns is None:
            _array_columns = array_columns
        return super()._parse_xml(xml_file, _array_columns=_array_columns)



"""
calibrator.py
====================================
Module for the calibrator functionality.
"""

from configparser import ConfigParser
from os.path import join
from pathlib import Path
from sys import stdout
from typing import Union

import numpy as np
import pandas as pd
from tqdm import tqdm

from gaiaxpy.config.paths import config_path, config_ini_file
from gaiaxpy.core.config import load_xpmerge_from_xml, load_xpsampling_from_xml
from gaiaxpy.core.generic_functions import cast_output, validate_wl_sampling, parse_band, format_sampled_output
from gaiaxpy.core.generic_variables import pbar_colour, pbar_units, pbar_message
from gaiaxpy.core.satellite import BANDS, BP_WL, RP_WL
from gaiaxpy.input_reader.input_reader import InputReader
from gaiaxpy.output.sampled_spectra_data import SampledSpectraData
from gaiaxpy.spectrum.sampled_basis_functions import SampledBasisFunctions
from gaiaxpy.spectrum.utils import get_covariance_matrix
from gaiaxpy.spectrum.xp_continuous_spectrum import XpContinuousSpectrum
from .external_instrument_model import ExternalInstrumentModel
from ..core.input_validator import validate_save_arguments
from ..spectrum.calibration_absolute_sampled_spectrum import CalibrationAbsoluteSampledSpectrum

__FUNCTION_KEY = 'calibrator'


def calibrate(input_object: Union[list, Path, pd.DataFrame, str], sampling: np.ndarray = None, truncation: bool = False,
              output_path: Union[Path, str] = '.', output_file: str = 'output_spectra', output_format: str = None,
              save_file: bool = True, with_correlation: bool = False, username: str = None, password: str = None) -> \
        (pd.DataFrame, np.ndarray):
    """
    Calibration utility: calibrates the input internally-calibrated continuously-represented mean spectra to the
    absolute system. An absolute spectrum sampled on a user-defined or default wavelength grid is created for each set
    of BP and RP input spectra. If either band is missing, the output spectrum will only cover the range covered by the
    available data.

    Args:
        input_object (list/Path/pd.DataFrame/str): Path to the file containing the mean spectra as downloaded from the
            Archive in their continuous representation, a list of sources ids (string or long), or a pandas DataFrame.
        sampling (ndarray): 1D array containing the desired sampling in absolute wavelengths [nm].
        truncation (bool): Toggle truncation of the set of bases. The level of truncation to be applied is defined by
            the recommended value in the input files.
        output_path (Path/str): Path where to save the output data.
        output_file (str): Name of the output file without extension (e.g. 'my_file').
        output_format (str): Desired output format. If no format is given, the output file format will be the same as
            the input file (e.g. 'csv').
        save_file (bool): Whether to save the output in a file. If false, output_format and output_file will be ignored.
        with_correlation (bool): Whether correlation information should be generated.
        username (str): Cosmos username, only suggested when input_object is a list or ADQL query.
        password (str): Cosmos password, only suggested when input_object is a list or ADQL query.

    Returns:
        (tuple): tuple containing:

            DataFrame: The values for all sampled absolute spectra.
            ndarray: The sampling used to calibrate the input spectra (user-provided or default).
    """
    return _calibrate(input_object, sampling, truncation, output_path, output_file, output_format, save_file,
                      with_correlation=with_correlation, username=username, password=password)


def _calibrate(input_object: Union[list, Path, str], sampling: np.ndarray = None, truncation: bool = False,
               output_path: Union[Path, str] = '.', output_file: str = 'output_spectra', output_format: str = None,
               save_file: bool = True, with_correlation: bool = False, username: str = None, password: str = None,
               bp_model: str = 'v375wi', rp_model: str = 'v142r', disable_info: bool = False) -> \
        (pd.DataFrame, np.ndarray):
    """
    Internal function of the calibration utility. Refer to "calibrate".

    Args:
        bp_model (str): The bp model.
        rp_model (str): The rp model.

    Returns:
        DataFrame: A list of all sampled absolute spectra.
        ndarray: The sampling used to calibrate the spectra.

    Raises:
        ValueError: If the sampling is out of the expected boundaries.
    """
    validate_wl_sampling(sampling)
    validate_save_arguments(_calibrate.__defaults__[3], output_file, _calibrate.__defaults__[4], output_format,
                            save_file)
    parsed_input_data, extension = InputReader(input_object, _calibrate, truncation=truncation,
                                               disable_info=disable_info, user=username, password=password).read()
    xp_design_matrices, xp_merge = __generate_xp_matrices_and_merge(__FUNCTION_KEY, sampling, bp_model, rp_model)
    spectra_df, positions = __create_spectra(parsed_input_data, truncation, xp_design_matrices, xp_merge,
                                             with_correlation=with_correlation, disable_info=disable_info)
    spectra_df = cast_output(spectra_df)
    output_data = SampledSpectraData(spectra_df, positions)
    output_data.save(save_file, output_path, output_file, output_format, extension)
    return spectra_df, positions


def __create_merge(xp: str, sampling: np.ndarray) -> np.ndarray:
    """
    Create the weight information on the input sampling grid.

    Args:
        xp (str): Band (either 'bp' or 'rp').
        sampling (ndarray): 1D array containing the sampling grid.

    Returns:
        ndarray: A numpy array containing an array with weights for the given band.
    """
    xp = parse_band(xp)
    wl_high = BP_WL.high
    wl_low = RP_WL.low
    return np.array([1.0 if wl < wl_low else 0.0 if wl > wl_high else (1.0 - (wl - wl_low) / (wl_high - wl_low)) for
                     wl in sampling]) \
        if xp == BANDS.bp else np.array([0.0 if wl < wl_low else 1.0 if wl > wl_high else (wl - wl_low) /
                                                                                          (wl_high - wl_low)
                                         for wl in sampling])


def __generate_xp_matrices_and_merge(label: str, sampling: np.ndarray, bp_model: str, rp_model: str) -> (dict, dict):
    """
    Generates the xp_design_matrices and xp_merge from the input parameters.

    Args:
        label (str): The label for the data.
        sampling (np.ndarray): The given sampling.
        bp_model (str): The bp model.
        rp_model (str): The rp model.

    Returns:
        tuple: A tuple containing two dictionaries, one for the xp_design_matrices and the other for the xp_merge.
    """

    def __get_file_for_xp(xp: str, key: str, _bp_model: str = bp_model, _rp_model: str = rp_model) -> str:
        """
        Retrieves the file for the specified xp band.

        Args:
            xp (str): The xp band, either 'bp' or 'rp'.
            key (str): The key for the file in the config file.
            _bp_model (str): The bp model.
            _rp_model (str): The rp model.

        Returns:
            str: The file path for the specified xp band.

        Raises:
            ValueError: If the xp band is not 'bp' or 'rp'.
        """
        xp = parse_band(xp)
        config_parser = ConfigParser()
        config_parser.read(config_ini_file)
        file_name = config_parser.get(label, key)
        model = _bp_model if xp == BANDS.bp else _rp_model
        return join(config_path, f"{file_name.replace('xp', xp).replace('model', model)}")

    if sampling is None:
        xp_sampling_grid, xp_merge = load_xpmerge_from_xml(bp_model=bp_model)
        xp_design_matrices = load_xpsampling_from_xml(bp_model=bp_model)
        xp_design_matrices = {xp: SampledBasisFunctions.from_design_matrix(xp_sampling_grid, xp_design_matrices[xp])
                              for xp in BANDS}
    else:
        xp_merge = {xp: __create_merge(xp, sampling) for xp in BANDS}
        xp_design_matrices = {xp: SampledBasisFunctions.from_external_instrument_model(
            sampling, xp_merge[xp], ExternalInstrumentModel.from_config_csv(
                __get_file_for_xp(xp, 'dispersion'), __get_file_for_xp(xp, 'response'),
                __get_file_for_xp(xp, 'bases'))) for xp in BANDS}
    return xp_design_matrices, xp_merge


def __create_spectra(parsed_input_data: pd.DataFrame, truncation: bool, design_matrices: dict,
                     merge: dict, with_correlation: bool = False, disable_info: bool = False):
    """
     Create a DataFrame of absolute sampled spectra for each source in the parsed mean spectra file.

     Args:
         parsed_input_data (DataFrame): DataFrame containing information for each source in the mean spectra file.
             This includes columns for both bands (although one band could be missing).
         truncation (bool): If True, the set of bases is truncated.
         design_matrices (dict): Dictionary containing 2D arrays of basis functions sampled on the pseudo-wavelength
             grid (either user-defined or default) for both bands.
         merge (dict): Dictionary containing arrays of weights for both bands. These define the contributions from each
             band to the joined absolute spectrum.
         with_correlation (bool, optional): If True, the covariance information is included in the resulting
             AbsoluteSampledSpectrum objects. Defaults to False.

     Returns:
         tuple:
             spectra_df (DataFrame): DataFrame of absolute sampled spectra, each represented as a dictionary with
                 attributes 'data_type' indicating the type of spectra and 'positions' indicating the sample positions.
             positions (ndarray): 1D array of the sample positions.
     """
    parsed_spectrum_file_dict = parsed_input_data.to_dict('records')
    spectra_series = pd.Series([_create_spectrum(row, truncation, design_matrices, merge,
                                                 with_correlation=with_correlation) for row in tqdm(
        parsed_spectrum_file_dict, desc=pbar_message[__FUNCTION_KEY], unit=pbar_units[__FUNCTION_KEY], leave=False,
        colour=pbar_colour, disable=disable_info, file=stdout)])
    return format_sampled_output(spectra_series, with_correlation=with_correlation)


def _create_spectrum(row, truncation, design_matrix, merge, with_correlation=False):
    """
    Create a single sampled absolute spectrum from the input continuously-represented mean spectrum and design matrix.

    Args:
        row (DataFrame): Single row in a DataFrame containing the entry for one source in the mean spectra file. This
            will include columns for both bands (although one could be missing).
        truncation (bool): Toggle truncation of the set of bases.
        design_matrix (ndarray): 2D array containing the basis functions sampled on the pseudo-wavelength grid (either
            user-defined or default).
        merge (dict): Dictionary containing an array of weights per BP and one for RP. These have one value per sample
            and define the contributions from BP and RP to the joined absolute spectrum.

    Returns:
        CalibrationAbsoluteSampledSpectrum: The absolute sampled spectrum with calibration behaviour.
    """
    source_id = row['source_id']
    continuous_dict = {band: XpContinuousSpectrum(source_id, band, row[f'{band}_coefficients'],
                                                  get_covariance_matrix(row, band), row[f'{band}_standard_deviation'])
                       for band in BANDS}
    recommended_truncation = {band: row[f'{band}_n_relevant_bases'] for band in BANDS} if truncation else dict()
    return CalibrationAbsoluteSampledSpectrum(source_id, continuous_dict, design_matrix, merge,
                                              truncation=recommended_truncation, with_correlation=with_correlation)



"""
xp_spectrum.py
====================================
Module to represent a BP/RP spectrum.
"""

from .generic_spectrum import Spectrum


class XpSpectrum(Spectrum):
    """
    A spectrum observed with one of the Gaia low-resolution photometers (Blue or Red Photometer, BP or RP).
    """

    def __init__(self, source_id, xp):
        """
        Initialise an XP spectrum.

        Args:
            source_id (int/str): Source identifier.
            xp (str): Gaia photometer, can be either 'bp' or 'rp'.
        """
        Spectrum.__init__(self, source_id)
        self.xp = xp

    def get_xp(self):
        """
        Get band.

        Returns:
            str: Photometer identifier. Either 'bp' or 'rp'.
        """
        return self.xp



"""
single_synthetic_photometry.py
====================================
Module to represent a synthetic photometry in a single photometric system.
"""

import warnings

import numpy as np

from .photometric_absolute_sampled_spectrum import PhotometricAbsoluteSampledSpectrum

# Ignore negative flux, handled in the code.
warnings.filterwarnings('ignore', category=RuntimeWarning)


class SingleSyntheticPhotometry(PhotometricAbsoluteSampledSpectrum):
    """
    Synthetic photometry derived from Gaia spectra in one photometric system.
    """

    def __init__(self, source_id, xp_spectra, sampled_bases, merge, truncation, photometric_system):
        """
        Initialise a synthetic photometry in a single photometric system.

        Args:
            source_id (str): Source identifier.
            xp_spectra (dict): A dictionary containing the BP and RP continuous spectra.
            sampled_bases (dict): The set of basis functions sampled onto the grid defining the resolution of the final
                sampled spectrum.
            merge (dict): The weighting factors for BP and RP sampled onto the grid defining the resolution of the final
                sampled spectrum.
            truncation (dict): The number of relevant bases per band.
            photometric_system (PhotometricSystem): The photometric system of the synthetic photometry.
        """
        PhotometricAbsoluteSampledSpectrum.__init__(self, source_id, xp_spectra, sampled_bases, merge,
                                                    truncation=truncation)
        self.photometric_system = photometric_system.value
        # Correct flux if necessary (regular Photometric systems return the original value)
        aux_flux = self.flux
        self.flux = self.photometric_system._correct_flux(aux_flux)
        # Correct the errors
        self.error = self.photometric_system._correct_error(aux_flux, self.error)
        # Magnitude is computed from self.flux which is the corrected flux
        self.mag = self._compute_mag()

    def _photometry_to_dict(self):
        """
        Represent the photometry as a dictionary.

        Returns:
            dict: A dictionary containing the source identifier and the synthetic photometry.
        """
        phot = {'source_id': self.source_id}
        mag = self._field_to_dict(self.mag, 'mag')
        flux = self._field_to_dict(self.flux, 'flux')
        error = self._field_to_dict(self.error, 'flux_error')
        return {**phot, **mag, **flux, **error}

    def _field_to_dict(self, field, name):
        """
        Represent the field as a dictionary.

        Returns:
            dict: A dictionary containing a particular field of the synthetic photometry.
        """
        bands = self.photometric_system.get_bands()
        values = field
        return {f'{name}_{band}': values[i] for i, band in enumerate(bands)}

    def _compute_mag(self):
        def flux_to_mag(flux, zero_point):
            """
            Convert flux to magnitude.

            Args:
                flux (float): Flux value.
                zero_point (float): Photometric zero-point.

            Returns:
                float: The magnitude.
            """
            if flux <= 0:
                return np.nan
            result = -2.5 * np.log10(flux) + zero_point
            return result

        return [flux_to_mag(flux, zero_point) for flux, zero_point in
                zip(self.flux, self.photometric_system.get_zero_points())]



"""
satellite.py
====================================
Module to hold satellite-related constants.
"""

from collections import namedtuple

# Satellite-related parameters
TELESCOPE_PUPIL_AREA = 0.7278

WLRange = namedtuple('WLRange', 'low high')

# Wavelength range covered by the BP spectrum
BP_WL = WLRange(330, 643)

# Wavelength range covered by the RP spectrum
RP_WL = WLRange(635, 1020)

# Bands
Bands = namedtuple('Bands', 'bp rp')
BANDS = Bands('bp', 'rp')



"""
parse_internal_continuous.py
====================================
Module to parse input files containing internally calibrated continuous spectra.
"""

import numpy as np
import pandas as pd
from fastavro import __version__ as fa_version
from hdfs import InsecureClient
from hdfs.ext.avro import AvroReader
from packaging import version
from requests.exceptions import ConnectionError

from gaiaxpy.core.generic_functions import array_to_symmetric_matrix, rename_with_required
from .cast import _cast
from .parse_generic import GenericParser
from .utils import _csv_to_avro_map, _get_from_dict
from ..core.custom_errors import SelectorNotImplementedError
from ..core.satellite import BANDS
from ..spectrum.utils import get_covariance_matrix

# Columns that contain arrays (as strings)
array_columns = ['bp_coefficients', 'bp_coefficient_errors', 'rp_coefficients', 'rp_coefficient_errors']
# Pairs of the form (matrix_size (N), values_to_put_in_matrix) for columns that contain matrices as strings
matrix_columns = [('bp_n_parameters', 'bp_coefficient_correlations'),
                  ('rp_n_parameters', 'rp_coefficient_correlations')]


class InternalContinuousParser(GenericParser):
    """
    Parser for internally calibrated continuous spectra.
    """

    def __init__(self, requested_columns=None, additional_columns=None, selector=None, **kwargs):
        super().__init__()
        self.additional_columns = dict() if additional_columns is None else additional_columns
        self.requested_columns = requested_columns
        self.selector = selector
        if kwargs:
            self.address = kwargs.get('address', None)
            self.port = kwargs.get('port', None)

    def _parse_csv(self, csv_file, _array_columns=None, _matrix_columns=None, _usecols=None):
        """
        Parse the input CSV file and store the result in a pandas DataFrame if it contains internally calibrated
            continuous spectra.

        Args:
            csv_file (str): Path to a CSV file.
            _array_columns (list): List of columns in the file that contain arrays as strings.
            _matrix_columns (list of tuples): List of tuples where the first element is the number of rows/columns of a
                square matrix which values are those contained in the second element of the tuple.
            _usecols (list): Columns to read.

        Returns:
            DataFrame: Pandas DataFrame representing the CSV file.
        """
        if self.selector is not None:
            raise SelectorNotImplementedError('E/CSV')
        if _matrix_columns is None:
            _matrix_columns = matrix_columns
        if _array_columns is None:
            _array_columns = array_columns
        _usecols = _usecols if _usecols else self.requested_columns
        df = super()._parse_csv(csv_file, _array_columns=_array_columns, _matrix_columns=_matrix_columns,
                                _usecols=_usecols)
        for band in BANDS:
            df[f'{band}_covariance_matrix'] = df.apply(get_covariance_matrix, axis=1, args=(band,))
        df = rename_with_required(df, self.additional_columns)
        return df

    def _parse_fits(self, fits_file, _array_columns=None, _matrix_columns=None, _usecols=None):
        """
        Parse the input FITS file and store the result in a pandas DataFrame if it contains internally calibrated
            continuous spectra.

        Args:
            fits_file (str): Path to a FITS file.
            _array_columns (list): List of columns in the file that contain arrays as strings.
            _matrix_columns (list of tuples): List of tuples where the first element is the number of rows/columns of a
                square matrix which values are those contained in the second element of the tuple.
            _usecols (list): Columns to read.

        Returns:
            DataFrame: Pandas DataFrame representing the FITS file.
        """
        if self.selector is not None:
            raise SelectorNotImplementedError('FITS')
        if _matrix_columns is None:
            _matrix_columns = matrix_columns
        if _array_columns is None:
            _array_columns = array_columns
        _usecols = _usecols if _usecols else self.requested_columns
        df = super()._parse_fits(fits_file, _array_columns=_array_columns, _matrix_columns=_matrix_columns,
                                 _usecols=_usecols)
        for band in BANDS:
            df[f'{band}_covariance_matrix'] = df.apply(get_covariance_matrix, axis=1, args=(band,))
        df = rename_with_required(df, self.additional_columns)
        return df

    def _parse_xml(self, xml_file, _array_columns=None, _matrix_columns=None, _usecols=None):
        """
        Parse the input XML file and store the result in a pandas DataFrame.

        Args:
            xml_file (str): Path to an XML file.
            _array_columns (list): List of columns in the file that contain arrays as strings.
            _matrix_columns (list of tuples): List of tuples where the first element is the number of rows/columns of a
                square matrix which values are those contained in the second element of the tuple.
            _usecols (list): Columns to read.

        Returns:
            DataFrame: A pandas DataFrame representing the XML file.
        """
        if self.selector is not None:
            raise SelectorNotImplementedError('XML')
        if _matrix_columns is None:
            _matrix_columns = matrix_columns
        if _array_columns is None:
            _array_columns = array_columns
        _usecols = _usecols if _usecols else self.requested_columns
        df = super()._parse_xml(xml_file, _array_columns=_array_columns, _matrix_columns=_matrix_columns,
                                _usecols=_usecols)
        for band in BANDS:
            df[f'{band}_covariance_matrix'] = df.apply(get_covariance_matrix, axis=1, args=(band,))
        df = rename_with_required(df, self.additional_columns)
        return df

    @staticmethod
    def __process_avro_record(record, additional_columns=None):
        _avro_keys_map = _csv_to_avro_map.copy()
        intersection_keys = [key for key in _avro_keys_map.keys() if key in additional_columns.keys()]
        if intersection_keys:
            raise ValueError('Additional columns will overwrite an already existing key. This is not allowed.'
                             f' Keys are: {",".join(intersection_keys)}')
        if additional_columns is not None:
            _avro_keys_map.update(additional_columns)
        return {key: np.array(_get_from_dict(record, _avro_keys_map[key])) if isinstance(
            _get_from_dict(record, _avro_keys_map[key]), list) else _get_from_dict(
            record, _avro_keys_map[key]) for key in _avro_keys_map.keys()}

    @staticmethod
    def __get_records_up_to_1_4_7(avro_file, additional_columns, selector, **kwargs):
        address = kwargs.get('address', None)
        if address:
            raise ValueError('HDFS access not implemented for fastavro versions older than 1.4.7.')
        from fastavro import reader
        f = open(avro_file, 'rb')
        avro_reader = reader(f)
        avro_reader = avro_reader if selector is None else filter(selector, avro_reader)
        record = avro_reader.next()
        while record:
            try:
                current_record = InternalContinuousParser.__process_avro_record(record, additional_columns)
                yield current_record
            except KeyError:
                raise KeyError("Keys in the input file don't match the expected ones.")
            try:
                record = avro_reader.next()
            except StopIteration:
                f.close()
                break

    @staticmethod
    def __get_records_later_than_1_4_7(avro_file, additional_columns, selector, **kwargs):
        def __yield_local_records(_avro_file):
            from fastavro import block_reader
            with open(_avro_file, 'rb') as fo:
                for block in block_reader(fo):
                    for rec in block:
                        yield rec

        def __yield_remote_records(_avro_file):
            client = InsecureClient(f'{address}:{port}')
            with AvroReader(client, _avro_file) as reader:
                for record in reader:
                    yield record

        address = kwargs.get('address', None)
        port = kwargs.get('port', None)
        records = __yield_remote_records(avro_file) if address else __yield_local_records(avro_file)
        records = records if selector is None else filter(selector, records)
        for record in records:
            yield InternalContinuousParser.__process_avro_record(record, additional_columns)

    def _parse_avro(self, avro_file):
        """
        Parse the input AVRO file and return the result as a Pandas DataFrame.

        Args:
            avro_file (str): Path to an AVRO file.

        Returns:
            DataFrame: Pandas DataFrame representing the AVRO file.
        """

        def __records_to_df(max_conn_retries=10, **_records_arguments):
            retries = 0
            while retries < max_conn_retries:
                try:
                    _df = pd.DataFrame(__get_records(**_records_arguments))
                    break
                except ConnectionError:
                    retries += 1
            else:
                raise ConnectionError(
                    f'Failed to connect to HDFS after {max_conn_retries} attempts for file {avro_file}.')
            return _df

        if version.parse(fa_version) <= version.parse('1.4.7'):
            __get_records = InternalContinuousParser.__get_records_up_to_1_4_7
        elif version.parse(fa_version) > version.parse('1.4.7'):
            __get_records = InternalContinuousParser.__get_records_later_than_1_4_7
        else:
            raise ValueError(f'Fastavro version {fa_version} may not have been parsed properly.')
        records_arguments = {
            'avro_file': avro_file,
            'additional_columns': self.additional_columns,
            'selector': self.selector
        }
        if hasattr(self, 'address') and hasattr(self, 'port'):
            records_arguments['address'] = self.address
            records_arguments['port'] = self.port
        df = __records_to_df(**records_arguments)
        # Pairs of the form (matrix_size (N), values_to_put_in_matrix)
        to_matrix_columns = [('bp_n_parameters', 'bp_coefficient_covariances'),
                             ('rp_n_parameters', 'rp_coefficient_covariances')]
        for size_column, values_column in to_matrix_columns:
            df[values_column] = df.apply(lambda row: array_to_symmetric_matrix(row[values_column], row[size_column]),
                                         axis=1)
        for band in BANDS:
            df[f'{band}_covariance_matrix'] = df.apply(get_covariance_matrix, axis=1, args=(band,))
        return _cast(df)



import subprocess
from os.path import splitext

from gaiaxpy.core.custom_errors import ExtensionNotImplementedError
from gaiaxpy.core.generic_functions import standardise_extension
from gaiaxpy.input_reader.file_reader import FileReader


def get_http_port():
    command = ['hdfs', 'getconf', '-confKey', 'dfs.namenode.http-address']
    process = subprocess.Popen(command, stdout=subprocess.PIPE)
    output, error = process.communicate()
    process.wait()
    if process.returncode != 0:
        raise Exception(f"Failed to execute command '{command}': {error}")
    return output.decode('utf-8').split(':')[1]


def split_cluster_path(file_path, expected_protocol='http', p_sep='://'):
    def process_address_and_port(_file_path, _protocol_split_index):
        expected_port = get_http_port()
        address_and_port = _file_path[:_protocol_split_index]
        split_index_port = address_and_port.find(':', len(f'{expected_protocol}{p_sep}'))
        if split_index_port != -1:
            _address, _port = address_and_port[:split_index_port], address_and_port[split_index_port:]
            if str(_port) != str(expected_port):
                _port = expected_port
            return _address, _port
        else:
            return address_and_port, expected_port

    def modify_protocol(_given_protocol, _expected_protocol, _file_path):
        if _given_protocol != _expected_protocol:
            return _expected_protocol + _file_path[len(_given_protocol):]
        return _file_path

    file_path = modify_protocol(file_path[:file_path.find(p_sep)], expected_protocol, file_path)
    protocol_split_index = file_path.find('/', len(f'{expected_protocol}{p_sep}'))
    if protocol_split_index == -1:
        raise ValueError('Input path has got a different format than expected.')
    else:
        address, port = process_address_and_port(file_path, protocol_split_index)
        file = file_path[protocol_split_index:]
        return address, file, port


class HDFSReader(FileReader):

    def __init__(self, file_parser_selector, file, truncation, additional_columns=None, selector=None,
                 disable_info=False):
        address, file_path, port = split_cluster_path(file)
        extension = standardise_extension(splitext(file_path)[1]).lower()
        if standardise_extension(splitext(file_path)[1]).lower() != 'avro':
            raise ExtensionNotImplementedError(extension)
        super().__init__(file_parser_selector, file_path, truncation, additional_columns, selector, disable_info,
                         address=address, port=port)



"""
multi_synthetic_photometry.py
====================================
Module to represent a synthetic photometry in multiple photometric systems.
"""
import re
from typing import List

import pandas as pd


def _flatten_list(lst):
    return [item for sublist in lst for item in sublist]


def _generate_variables(photometries):
    # A photometry corresponds to the photometry of a single source id in all the different systems
    mags = [[photometry.mag for photometry in multi_photometry] for multi_photometry in photometries]
    fluxes = [[photometry.flux for photometry in multi_photometry] for multi_photometry in photometries]
    flux_errors = [[photometry.error for photometry in multi_photometry] for multi_photometry in photometries]
    return mags, fluxes, flux_errors


def _get_source_ids(photometries):
    return [photometry[0].source_id for photometry in photometries]


class MultiSyntheticPhotometry(object):
    """
    Synthetic photometry derived from Gaia spectra in multiple photometric systems.
    """

    def __init__(self, photometric_system, photometries):
        """
        Initialise a synthetic photometry in multiple photometric systems.
        """
        self.photometric_system = photometric_system
        # One element per source_id, each sub element corresponds to one photometric_system
        self.photometries = photometries
        self.source_ids = _get_source_ids(photometries)
        self.mags, self.fluxes, self.errors = _generate_variables(photometries)

    def _generate_output_df(self):
        def _get_column_sublist(_label: str, _df: pd.DataFrame) -> List:
            """
            Get the columns in a DataFrame that correspond to a particular system.

            Args:
                _label: A system label.
                _df: A DataFrame containing photometry data.

            Returns:
                A list of column names.
            """
            _pattern = fr'{_label}_(flux_error|flux|mag)_[A-Za-z0-9]+'
            return list(filter(lambda col: re.match(_pattern, col), _df.columns))

        photometries_df = self._photometries_to_df()
        # Reorder DataFrame columns
        phot_system_labels = [phot_system.get_system_label() for phot_system in self.photometric_system]
        reordered_columns = ['source_id'] + [col for label in phot_system_labels for col in _get_column_sublist(
            label, photometries_df)]
        photometries_df = photometries_df[reordered_columns]
        return photometries_df

    def _photometries_to_df(self):
        list_of_dicts = []
        for source_id, mags, fluxes, errors in zip(self.source_ids, self.mags, self.fluxes, self.errors):
            phot = {'source_id': source_id}
            mag = self._field_to_dict(mags, 'mag')
            flux = self._field_to_dict(fluxes, 'flux')
            error = self._field_to_dict(errors, 'flux_error')
            list_of_dicts.append({**phot, **mag, **flux, **error})
        return pd.DataFrame(list_of_dicts)

    def _field_to_dict(self, values, name):
        def _build_dict_keys(_system_bands_tuples, _name):
            dict_keys = []
            for system_bands in _system_bands_tuples:
                system = system_bands[0]
                bands = system_bands[1]
                for band in bands:
                    dict_keys.append(f'{system}_{_name}_{band}')
            return dict_keys

        system_bands_tuples = [(phot_system.get_system_label(), phot_system.get_bands()) for phot_system in
                               self.photometric_system]
        keys_list = _build_dict_keys(system_bands_tuples, name)
        values = _flatten_list(values)
        return {key: value for key, value in zip(keys_list, values)}



"""
output_data.py
====================================
Module to represent generic output data.
"""
from ast import literal_eval
from os.path import dirname, abspath, join

from numpy import set_printoptions

from gaiaxpy.core.generic_functions import standardise_extension
from gaiaxpy.file_parser.parse_generic import InvalidExtensionError

set_printoptions(legacy='1.21')


def _initialise_header():
    return ["# %ECSV 1.0", "# ---", "# delimiter: ','", "# datatype:"]


def _load_header_dict():
    current_path = dirname(abspath(__file__))
    header_dictionary_path = join(current_path, 'ecsv_headers', 'headers_dict.txt')
    with open(header_dictionary_path) as f:
        data = f.read()
    # Load header dictionary
    header_dict = literal_eval(data)
    return header_dict


def _build_regular_header(columns):
    header_dict = _load_header_dict()
    header = _initialise_header()
    for column in columns:
        current_column = header_dict[column]
        header.append('# -')
        header.append(f'#   name: {column}')
        header.append(f'#   datatype: {current_column["datatype"]}')
        if current_column.get('subtype'):
            header.append(f'#   subtype: {current_column["subtype"]}')
        header.append(f'#   description: {current_column["description"]}')
    return '\n'.join(header) + '\n'


class OutputData(object):

    def __init__(self, data, positions):
        self.data = data.copy()
        self.positions = positions

    def save(self, save_file, output_path, output_file, output_format, extension):
        """
        Save the output data.

        Args:
            save_file (bool): Whether to save the file or not.
            output_path (str): Path where to save the file.
            output_file (str): Name of the output file.
            output_format (str): Format of the output file.
            extension (str): Format of the original input file.
        """
        if save_file:
            if output_file is None:
                raise ValueError('The parameter output_file cannot be None.')
            if output_format is None:
                output_format = extension
            print('Saving file...', end='\r')
            output_format = standardise_extension(output_format)
            if output_format == 'avro':
                self._save_avro(output_path, output_file)
            elif output_format == 'csv':
                self._save_csv(output_path, output_file)
            elif output_format == 'ecsv':
                self._save_ecsv(output_path, output_file)
            elif output_format == 'fits':
                self._save_fits(output_path, output_file)
            elif output_format == 'xml':
                self._save_xml(output_path, output_file)
            else:
                raise InvalidExtensionError()
            print(f"Done! Output saved to path: {join(output_path, output_file + '.' + output_format)}", end='\r')

    def _save_avro(self, output_path, output_file):
        raise NotImplementedError()

    def _save_csv(self, output_path, output_file):
        raise NotImplementedError()

    def _save_ecsv(self, output_path, output_file):
        raise NotImplementedError()

    def _save_fits(self, output_path, output_file):
        raise NotImplementedError()

    def _save_xml(self, output_path, output_file):
        raise NotImplementedError()



from gaiaxpy.core.server import data_release, gaia_server
from .archive_reader import ArchiveReader
from .dataframe_reader import DataFrameReader
from ..core.custom_errors import SelectorNotImplementedError

not_supported_functions = ['apply_colour_equation', 'apply_error_correction']


def extremes_are_enclosing(first_row, column):
    if first_row[column][0] == '[' and first_row[column][-1] == ']':
        return True
    elif first_row[column][0] == '(' and first_row[column][-1] == ')':
        return True
    else:
        return False


class ListReader(ArchiveReader):

    def __init__(self, content, function, truncation, user, password, additional_columns=None, selector=None,
                 disable_info=False):
        if selector is not None:
            raise SelectorNotImplementedError('List')
        if additional_columns is None:
            additional_columns = dict()
        super(ListReader, self).__init__(function, truncation, user, password, additional_columns=additional_columns,
                                         disable_info=disable_info)
        if content:
            self.content = content
        else:
            raise ValueError('Input list cannot be empty.')
        self.disable_info = disable_info

    def read(self, _data_release=data_release):
        # Import only when read is called, as this module is dependent on the Archive’s availability
        from astroquery.gaia import GaiaClass

        sources = self.content
        function_name = self.function.__name__
        if function_name in not_supported_functions:
            raise ValueError(f'Function {function_name} does not support receiving a list as input.')
        # Connect to geapre
        gaia = GaiaClass(gaia_tap_server=gaia_server, gaia_data_server=gaia_server)
        self._login(gaia)
        # ADQL query
        if not self.disable_info:
            self.show_info_msg()
        result = gaia.load_data(ids=sources, format='csv', data_release=_data_release, data_structure='raw',
                                retrieval_type='XP_CONTINUOUS', avoid_datatype_check=True)
        try:
            continuous_key = [key for key in result.keys() if 'continuous' in key.lower()][0]
            data = result[continuous_key][0].to_pandas()
        except (KeyError, IndexError):
            raise ValueError('No continuous BP/RP data found for the given sources.')
        if not self.disable_info:
            self.show_info_msg(done=True)
        return DataFrameReader(data, function_name, self.truncation, additional_columns=self.additional_columns,
                               disable_info=True).read()



"""
xp_filter_system_colour_equation.py
====================================
Module that implements the colour equation functionality.
"""

import math
from ast import literal_eval
from configparser import ConfigParser
from os import listdir
from pathlib import Path
from sys import stdout
from typing import Union, Optional

import numpy as np
import pandas as pd
from numpy import poly1d
from tqdm import tqdm

from gaiaxpy.config.paths import filters_path
from gaiaxpy.core.generic_functions import cast_output
from gaiaxpy.core.generic_variables import pbar_colour, pbar_units, pbar_message
from gaiaxpy.core.input_validator import validate_save_arguments
from gaiaxpy.generator.photometric_system import PhotometricSystem
from gaiaxpy.input_reader.dataframe_reader import DataFrameReader
from gaiaxpy.output.photometry_data import PhotometryData

colour_eq_dir = Path(filters_path, '..', 'colour_eq_files')


def __raise_key_error(key):
    """
    Raise a KeyError with a specific error message.

    Args:
        key (str): The missing key that triggered the error.

    Raises:
        KeyError: A KeyError with a specific error message indicating the missing key.
    """
    raise KeyError(f'Required column {key} is not present in input data.')


def __compute_mag_error(data, band, system_label):
    """
    Compute magnitude error based on flux error.

    Args:
        data (dict): A dictionary containing the input data.
        band (str): The band to use.
        system_label (str): The system being used.

    Returns:
        float: The magnitude error.
    """
    return 2.5 * data[f'{system_label}_flux_error_{band}'] / (data[f'{system_label}_flux_{band}'] * math.log(10))


def __fill_systems_details(systems_to_correct):
    """
    Get the details of the systems that will be corrected.

    Args:
        systems_to_correct (list): A list containing the systems to be corrected.

    Returns:
        dict: A dictionary containing the details of the given systems.
    """
    systems_details = dict()
    for system in systems_to_correct:
        label = system.get_system_label()
        systems_details[label] = dict()
        # Get bands and zero points
        systems_details[label]['bands_zp'] = dict(zip(system.get_bands(), system.get_zero_points()))
        # Load ini file
        config_parser = ConfigParser()
        config_parser.read(Path(colour_eq_dir, f'{label}_colour_eq.ini'))
        systems_details[label]['filter'] = config_parser.get(label, 'FILTER')  # The filter to be corrected (string)
        systems_details[label]['colour_index'] = config_parser.get(label, 'COLOUR_INDEX')  # The colour index (string)
        # The colour equation (PolynomialFunction)
        # Reverse, coefficients were originally defined for the Java polyfunction
        coefficients = list(literal_eval(config_parser.get(label, 'POLY_COEFFICIENTS')))[::-1]
        polyfunc = poly1d(coefficients)
        systems_details[label]['polyfunc'] = polyfunc
        systems_details[label]['derivative'] = polyfunc.deriv()  # Colour equation derivative (UnivariateFunction)
        colour_range = config_parser.get(label, 'COLOUR_RANGE')  # Colour range for the correction (double[])
        systems_details[label]['colour_range'] = literal_eval(colour_range)
    return systems_details


def __create_rows(single_system_df, system, colour_band_0, colour_band_1, systems_details):
    """
    Create output rows for a systems.
    """
    new_system_rows = [__generate_output_row(row, system, colour_band_0, colour_band_1, systems_details) for row in
                       single_system_df.to_dict('records')]
    return pd.DataFrame(new_system_rows)


def _generate_output_df(input_synthetic_photometry, systems_details, disable_info=False):
    __FUNCTION_KEY = 'colour_eq'
    synth_phot_df = input_synthetic_photometry.copy()
    column_names = synth_phot_df.columns
    # Extract columns corresponding to one system
    system_keys = systems_details.keys()
    for label in tqdm(system_keys, desc=pbar_message[__FUNCTION_KEY], total=len(system_keys),
                      unit=pbar_units[__FUNCTION_KEY], colour=pbar_colour, leave=False, disable=disable_info,
                      file=stdout):
        filter_to_correct = systems_details[label]['filter']
        colour_band_0, colour_band_1 = _get_colour_bands(systems_details[label]['colour_index'])
        system_columns_with_colour = [column for column in column_names if column.startswith(f'{label}_') and
                                      column.endswith((f'_{filter_to_correct}', f'_{colour_band_0}',
                                                       f'_{colour_band_1}'))]
        single_system_df = synth_phot_df[system_columns_with_colour]
        corrected_system_df = __create_rows(single_system_df, label, colour_band_0, colour_band_1, systems_details)
        columns_to_correct = corrected_system_df.columns
        synth_phot_df[columns_to_correct] = corrected_system_df[columns_to_correct]
    return synth_phot_df


def __generate_output_row(row, system_label, colour_band_0, colour_band_1, systems_details):
    filter_to_correct = systems_details[system_label]['filter']
    mag = row[f'{system_label}_mag_{filter_to_correct}']
    mag_err = __compute_mag_error(row, filter_to_correct, system_label)
    mag_colour_0 = row[f'{system_label}_mag_{colour_band_0}']
    mag_colour_1 = row[f'{system_label}_mag_{colour_band_1}']
    colour = mag_colour_0 - mag_colour_1
    # Output corrected magnitude
    corrected_magnitude = mag + _get_correction(systems_details, colour, system_label)
    # Propagated colour error
    mag_err_1 = __compute_mag_error(row, colour_band_0, system_label)
    mag_err_2 = __compute_mag_error(row, colour_band_1, system_label)
    colour_err = math.sqrt(mag_err_1 ** 2 + mag_err_2 ** 2)
    correction_err = colour_err * abs(systems_details[system_label]['derivative'](colour))
    # Total error on corrected magnitude
    out_err = math.sqrt(mag_err ** 2 + correction_err ** 2)
    new_row = dict()
    new_row[f'{system_label}_mag_{filter_to_correct}'] = corrected_magnitude
    zp = systems_details[system_label]['bands_zp'][filter_to_correct]
    out_flux = 10 ** (-0.4 * (corrected_magnitude - zp))
    out_flux_err = out_err * out_flux * math.log(10) / 2.5
    new_row[f'{system_label}_flux_{filter_to_correct}'] = out_flux
    new_row[f'{system_label}_flux_error_{filter_to_correct}'] = out_flux_err
    return new_row


def _get_colour_bands(colour_index):
    """
    Split the bands of a colour index.

    Args:
        colour_index (str): Colour index containing two bands separated by a hyphen (e.g. U-V).

    Returns:
        tuple: Tuple of two elements containing the bands for the colour index as strings.
    """
    colour_band_0, colour_band_1 = colour_index.split('-')
    return colour_band_0, colour_band_1


def _set_colour_limit(colour, colour_range):
    """
    Set the colour limit to either the minimum or maximum value in the given colour range.

    Args:
        colour (float): The colour value.
        colour_range (list): A list of two values representing the minimum and maximum allowed values for the colour.

    Returns:
        float: The minimum or maximum value in the colour range, whichever is closest to the input `colour` value.

    Raises:
        ValueError: If the input `colour` value is not within the specified `colour_range`.
    """
    if colour < min(colour_range):
        colour_limit = min(colour_range)
    elif colour > max(colour_range):
        colour_limit = max(colour_range)
    else:
        raise ValueError(
            'The condition for one of the previous statements must be True. Error in colour or colour_range.')
    return colour_limit


def _generate_polynomial(colour, colour_limit, system_polyfunc):
    v = system_polyfunc(colour_limit)
    m = system_polyfunc.deriv()(colour_limit)
    return v + m * (colour - colour_limit)


# Replaces Java 'contains'
def _is_in_range(colour: float, colour_range: list):
    return min(colour_range) <= colour <= max(colour_range)


def _get_correction(systems_details: dict, colour: float, system_label: str):
    colour_range = systems_details[system_label]['colour_range']
    polyfunc = systems_details[system_label]['polyfunc']
    # Check whether colour is nan
    if math.isnan(colour):
        return np.nan
    # Check if in range after being sure that colour is not nan
    in_range = _is_in_range(colour, colour_range)
    if in_range:
        return polyfunc(colour)
    elif not in_range:
        colour_limit = _set_colour_limit(colour, colour_range)
        return _generate_polynomial(colour, colour_limit, polyfunc)
    else:
        raise ValueError('At least one variable does not comply with any of the previous statements.')


def __get_systems_to_correct(systems: Union[list, PhotometricSystem]) -> list:
    systems = [systems] if isinstance(systems, PhotometricSystem) else systems
    correctable_labels = [filename.split('_')[0] for filename in listdir(colour_eq_dir)]
    correctable_systems = [system for system in systems if system.get_system_label() in correctable_labels]
    return correctable_systems


def apply_colour_equation(input_synthetic_photometry: pd.DataFrame,
                          photometric_system: Union[list, PhotometricSystem] = None, output_path: str = '.',
                          output_file: str = 'corrected_photometry', output_format: Optional[str] = None,
                          save_file: bool = True):
    """
    Apply the available colour correction to the input photometric system(s).

    Args:
        input_synthetic_photometry (DataFrame): Input photometry as returned by GaiaXPy generator.
        photometric_system (PhotometricSystem, list of PhotometricSystem): The photometric systems over which to apply
        the equations.
        output_path (str): The path of the output file.
        output_file (str): Name of the output file without extension (e.g. 'my_file').
        output_format (str): Desired output format. If no format is given, the output file format will be the same as
            the input file (e.g. 'csv').
        save_file (bool): Whether to save the output file.

    Returns:
        DataFrame: The input photometry with colour equations applied if it corresponds.
    """
    return _apply_colour_equation(input_synthetic_photometry=input_synthetic_photometry,
                                  photometric_system=photometric_system, output_path=output_path,
                                  output_file=output_file, output_format=output_format, save_file=save_file)


def _apply_colour_equation(input_synthetic_photometry: pd.DataFrame,
                           photometric_system: Union[list, PhotometricSystem] = None, output_path: str = '.',
                           output_file: str = 'corrected_photometry', output_format: Optional[str] = None,
                           save_file: bool = True, disable_info=False):
    function = apply_colour_equation
    validate_save_arguments(function.__defaults__[2], output_file, function.__defaults__[3], output_format, save_file)
    input_synthetic_photometry, extension = DataFrameReader(input_synthetic_photometry, function, False,
                                                            disable_info=disable_info).read()
    systems_to_correct = __get_systems_to_correct(photometric_system)
    systems_details = __fill_systems_details(systems_to_correct)
    output_df = _generate_output_df(input_synthetic_photometry, systems_details, disable_info=disable_info)
    output_data = PhotometryData(output_df)
    output_data.data = cast_output(output_data)
    output_data.save(save_file, output_path, output_file, output_format, extension)
    return output_data.data



"""
sampled_spectra_data.py
====================================
Module to represent a dataframe of sampled spectra.
"""

import warnings
from os.path import join
from pathlib import Path

import pandas as pd
from astropy.io import fits
from astropy.io.votable.tree import Field, Param, Resource, VOTableFile
from astropy.units import UnitsWarning
from fastavro import parse_schema, writer
from fastavro.validation import validate_many
from numpy import ndarray

from .output_data import OutputData
from .utils import (_add_ecsv_header, _array_to_standard, _build_ecsv_header, _generate_fits_header,
                    _get_sampling_dict, _load_header_dict, _get_col_subtype_len)

try:
    from astropy.io.votable.tree import TableElement as ATable
except ImportError:
    from astropy.io.votable.tree import Table as ATable


class SampledSpectraData(OutputData):

    def __init__(self, data, positions):
        super().__init__(data, positions)

    def _save_avro(self, output_path, output_file):
        """
        Save the output spectra in AVRO format.

        Args:
            output_path (str): Path where to save the file.
            output_file (str): Name of the output file.
        """

        def _save_avro_sampling(_positions, _output_path, _output_file):
            """
            Save the sampling in a separate avro file.

            Args:
                _positions (list): Sampling positions.
                _output_path (str): Path where to store the output file.
                _output_file (str): Name of the output file.
            """
            schema = {'doc': 'Output sampling.', 'name': 'Sampling', 'namespace': 'sampling', 'type': 'record',
                      'fields': [{'name': 'pos', 'type': 'string'}, ], }
            # Must be an iterable
            sampling = [_get_sampling_dict(_positions)]
            # Sampling field to string
            sampling[0]['pos'] = str(sampling[0]['pos'])
            # Validate that records match the schema
            validate_many(sampling, schema)
            _parsed_schema = parse_schema(schema)
            with open(join(_output_path, f'{_output_file}_sampling.avro'), 'wb') as _output:
                writer(_output, _parsed_schema, sampling)

        def _generate_avro_schema(_spectra_dicts):
            """
            Generate the AVRO schema required to store the output.

            Args:
                _spectra_dicts (list): A list of dictionaries containing spectra.

            Returns:
                dict: A dictionary containing the parsed schema that matches the input.
                list: A list of dictionaries with the modified input spectra according to the valid AVRO types.
            """
            field_to_type = {'source_id': 'long', 'xp': 'string', 'flux': 'string', 'flux_error': 'string',
                             'correlation': 'string', 'standard_deviation': 'float'}

            def build_field(keys):
                return [{'name': key, 'type': field_to_type[key]} for key in keys]

            schema = {'doc': 'Spectrum output.', 'name': 'Spectra', 'namespace': 'spectrum', 'type': 'record',
                      'fields': build_field(_spectra_dicts[0].keys()), }
            # Spectrum fields to string
            for spectrum in _spectra_dicts:
                for field, _type in field_to_type.items():
                    if _type == 'string' and field in spectrum.keys() and not field == 'xp':
                        spectrum[field] = str(tuple(spectrum[field]))
            # Validate that records match the schema
            validate_many(_spectra_dicts, schema)
            return parse_schema(schema), _spectra_dicts

        data = self.data
        positions = self.positions
        Path(output_path).mkdir(parents=True, exist_ok=True)
        _save_avro_sampling(positions, output_path, output_file)
        # List with one dictionary per source
        spectra_dicts = data.to_dict('records')
        parsed_schema, spectra_dicts = _generate_avro_schema(spectra_dicts)
        output_path = join(output_path, f'{output_file}.avro')
        with open(output_path, 'wb') as output:
            writer(output, parsed_schema, spectra_dicts)

    def _save_csv(self, output_path, output_file):
        """
        Save the output spectra in CSV format.

        Args:
            output_path (str): Path where to save the file.
            output_file (str): Name of the output file.
        """
        data = self.data
        positions = self.positions
        modified_data = data.map(lambda x: _array_to_standard(x) if isinstance(x, ndarray) else x)
        Path(output_path).mkdir(parents=True, exist_ok=True)
        modified_data.to_csv(join(output_path, f'{output_file}.csv'), index=False)
        # Assume the sampling is the same for all spectra
        pos = [str(_array_to_standard(positions))]
        sampling_df = pd.DataFrame({'pos': pos})
        sampling_df.to_csv(join(output_path, f'{output_file}_sampling.csv'), index=False)

    def _save_ecsv(self, output_path, output_file):
        """
        Save the output spectra in ECSV format.

        Args:
            output_path (str): Path where to save the file.
            output_file (str): Name of the output file.
        """
        data = self.data
        positions = self.positions
        modified_data = data.map(lambda x: _array_to_standard(x, 'ecsv') if isinstance(x, ndarray) else x)
        header_lines = _build_ecsv_header(modified_data, positions)
        Path(output_path).mkdir(parents=True, exist_ok=True)
        modified_data.to_csv(join(output_path, f'{output_file}.ecsv'), index=False)
        _add_ecsv_header(header_lines, output_path, output_file)

    def _save_fits(self, output_path, output_file):
        """
        Save the output data in FITS format.

        Args:
            output_path (str): Path where to save the file.
            output_file (str): Name of the output file.
        """
        warnings.filterwarnings('ignore', category=UnitsWarning)

        def _flux_contains_none(arrays):
            return any(arr is None for arr in arrays['flux'])

        data = self.data
        positions = self.positions
        # Create a list of HDUs
        hdu_list = list()
        # create a header to include the sampling
        hdr = fits.Header()
        primary_hdu = fits.PrimaryHDU(header=hdr)
        hdu_list.append(primary_hdu)
        # Create a dictionary to hold all the data
        output_by_column_dict = data.reset_index().to_dict(orient='list')
        # Remove index from output dict
        output_by_column_dict.pop('index', None)
        spectra_keys = output_by_column_dict.keys()
        data_type = data.attrs['data_type']
        units_dict = data_type.get_units()
        pos_len = len(positions)
        contains_none = _flux_contains_none(output_by_column_dict)
        # D: double precision float
        flux_format = 'PD()' if contains_none else f'{pos_len}D'
        # E: single precision float
        flux_error_format = 'PE()' if contains_none else f'{pos_len}E'
        aux_corr = data.get('correlation')
        correlation_format = ''
        if aux_corr is not None:
            correlation_format = 'PD()' if contains_none else f"{_get_col_subtype_len(data, 'correlation')}D"
        # Define formats for each type according to FITS
        column_formats = {'source_id': 'K', 'xp': '2A', 'flux': flux_format, 'flux_error': flux_error_format,
                          'correlation': correlation_format, 'standard_deviation': 'E'}
        columns = [
            fits.Column(name=key, array=[value if value is not None else [] for value in output_by_column_dict[key]],
                        format=column_formats[key], unit=units_dict.get(key, '')) for key in spectra_keys]
        header = _generate_fits_header(data, column_formats)
        header['Sampling'] = str(tuple(positions))
        hdu = fits.BinTableHDU.from_columns(columns, header=header)
        hdu_list.append(hdu)
        # Put all HDUs together
        hdul = fits.HDUList(hdu_list)
        # Write the file and replace it if it already exists
        Path(output_path).mkdir(parents=True, exist_ok=True)
        output_path = join(output_path, f'{output_file}.fits')
        hdul.writeto(output_path, overwrite=True)

    def _save_xml(self, output_path, output_file):
        """
        Save the output spectra in XML/VOTABLE format.

        Args:
            output_path (str): Path where to save the file.
            output_file (str): Name of the output file.
        """
        warnings.filterwarnings('ignore', category=UnitsWarning)

        def _create_params(_votable, sampling, data_type):
            column = 'sampling'
            unit = data_type.get_units().get('pos', '')
            return [Param(_votable, name=column, ID=f'_{column}', ucd='em.wl', datatype='double', arraysize='*',
                          unit=unit, value=list(sampling))]

        def _create_fields(_votable, _spectra_df):
            _spectra_flux_len = _get_col_subtype_len(_spectra_df, 'flux')
            _spectra_flux_error_len = _get_col_subtype_len(_spectra_df, 'flux_error')
            len_flux = str(_spectra_flux_len)
            len_error = str(_spectra_flux_error_len)
            len_correlation = str(
                len(_spectra_df['correlation'].iloc[0])) if 'correlation' in _spectra_df.columns else ''
            fields_datatypes = {'source_id': 'long', 'xp': 'char', 'flux': 'double', 'flux_error': 'float',
                                'correlation': 'double', 'standard_deviation': 'float'}
            fields_array_size = {'source_id': '', 'xp': '2', 'flux': len_flux, 'flux_error': len_error,
                                 'correlation': len_correlation, 'standard_deviation': ''}
            fields_id = {key: f'_{key}' for key in ['source_id', 'xp', 'flux', 'flux_error', 'correlation']}
            fields_id.update({'source_id': None})
            header_dict = _load_header_dict()
            data_type = _spectra_df.attrs['data_type']
            units_dict = data_type.get_units()
            _fields = [Field(_votable, name=column, ucd=header_dict.get(column, dict()).get('meta', ''),
                             ID=fields_id[column], datatype=fields_datatypes[column],
                             arraysize=fields_array_size[column], unit=units_dict.get(column, None)) if
                       fields_array_size[column] != '' else Field(_votable, name=column,
                                                                  datatype=fields_datatypes[column],
                                                                  ucd=header_dict.get(column, dict()).get('meta', ''),
                                                                  unit=units_dict.get(column, None))
                       for column in _spectra_df.columns]
            for _field in _fields:
                _field.description = header_dict.get(_field.name, dict()).get('description', '')
            return _fields

        spectra_df = self.data
        positions = list(self.positions)
        # Create a new VOTable file
        votable = VOTableFile()
        # Add a resource
        resource = Resource()
        votable.resources.append(resource)
        # Add a table for the spectra (and add the sampling as metadata)
        spectra_table = ATable(votable)
        resource.tables.append(spectra_table)
        # Add sampling as param
        params = _create_params(votable, positions, spectra_df.attrs['data_type'])
        spectra_table.params.extend(params)
        # Add spectrum fields
        fields = _create_fields(votable, spectra_df)
        spectra_table.fields.extend(fields)
        # Create the record arrays, with the given number of rows
        spectra_table.create_arrays(len(spectra_df))
        for index, row in enumerate(spectra_df.to_dict('records')):
            spectra_table.array[index] = tuple([row[column] for column in spectra_df.columns])
        # Write to a file
        Path(output_path).mkdir(parents=True, exist_ok=True)
        output_path = join(output_path, f'{output_file}.xml')
        votable.to_xml(output_path)
